static const char _rcsid_[]="$Id: xio-ipapp.c,v 1.13 2003/02/15 14:26:03 gerhard Exp $";
/* Copyright Gerhard Rieger 2001, 2002 */
/* Published under the GNU General Public License V.2, see file COPYING */

/* this file contains the source for TCP and UDP related options */

#include "xiosysincludes.h"
#include "xioopen.h"

#include "xio-socket.h"
#include "xio-listen.h"
#include "xio-ip6.h"
#include "xio-ipapp.h"

#if WITH_TCP || WITH_UDP

const struct optdesc opt_sourceport = { "sourceport", "sp",       OPT_SOURCEPORT,  GROUP_IPAPP,     PH_LATE,TYPE_2BYTE,	OFUNC_SPEC };
/*const struct optdesc opt_port = { "port",  NULL,    OPT_PORT,        GROUP_IPAPP, PH_BIND,    TYPE_USHORT,	OFUNC_SPEC };*/
const struct optdesc opt_lowport = { "lowport", NULL, OPT_LOWPORT, GROUP_IPAPP, PH_LATE, TYPE_BOOL, OFUNC_SPEC };

#if WITH_IP4
/* we expect the form "host:port" */
int xioopen_ip4app_connect(char *a1, int rw, xiofile_t *fd,
			 unsigned groups, int socktype,
			 int ipproto, void *protname) {
   struct opt *opts = NULL;
   char *a2, *a3;
   int result;

   if (a2 = strchr(a1, ':')) {
      *a2++ = '\0';
   } else {
      Error1("xioopen_ip4app_connect(\"%s\", ...): missing parameter(s)", a1);
      return -1;
   }

   if (a3 = strchr(a2, ',')) {
      *a3++ = '\0';
   }
   if (parseopts(a3, groups, &opts) < 0) {
      return STAT_NORETRY;
   }

   if ((result =
	_xioopen_ip4app_connect(a1, a2, fd,
				socktype, ipproto, protname, opts)) < 0) {
      return result;
   }

   if ((result = _xio_openlate(&fd->stream, opts)) < 0) {
      return result;
   }
   return 0;
}

int _xioopen_ip4app_connect(const char *hostname, const char *portname,
			    xiofile_t *xfd,
			    int socktype, int ipproto, void *protname,
			    struct opt *opts) {
   bool needbind = false;
   struct sockaddr_in us;
   struct sockaddr_in them;
   byte2_t port;
   bool lowport = false;
   int result;

   /* initialization of local vars */
   socket_in_init(&them);
   socket_in_init(&us);

   if ((result =
	_parseip4addr(hostname, portname, &them, (char *)protname, 1, 1))
       < 0)
      return result;

   applyopts(-1, opts, PH_INIT);

   if (retropt_2bytes(opts, OPT_SOURCEPORT, &port) >= 0) {
      us.sin_port = htons(port);
      needbind = true;
   }

   retropt_bool(opts, OPT_LOWPORT, &lowport);

   /* 3 means: IP address AND port accepted */
   if (retropt_bind(opts, AF_INET, (char *)protname, (struct sockaddr *)&us, 3) >= 0)
      needbind = true;

   retropt_int(opts, OPT_SO_TYPE, &socktype);

   if ((result =
	_xioopen_connect(&xfd->stream,
		      needbind?(struct sockaddr *)&us:NULL, sizeof(us),
		      (struct sockaddr *)&them, sizeof(struct sockaddr_in),
		      opts, PF_INET, socktype, ipproto, lowport)) < 0) {
      return result;
   }
   return 0;
}
#endif /* WITH_IP4 */


#if WITH_IP4 && WITH_TCP && WITH_LISTEN
/* we expect the form: port */
/* currently only used for TCP4 */
int xioopen_ip4app_listen(char *a1, int rw, xiofile_t *fd,
			 unsigned groups, int socktype,
			 int ipproto, void *protname) {
   char *a2;
   struct opt *opts = NULL;
   struct sockaddr_in us;
   struct hostent *usaddr;
   char *bindname;
   int result;

   if (a2 = strchr(a1, ','))  *a2++ = '\0';
   if (parseopts(a2, groups, &opts) < 0)
      return STAT_NORETRY;
   applyopts(-1, opts, PH_INIT);

   socket_in_init(&us);
   us.sin_port = parseport(a1, (char *)protname);

   if (retropt_string(opts, OPT_BIND, &bindname) >= 0) {
      usaddr = Gethostbyname(bindname);
      if (usaddr == NULL) {
	 Error1("cannot resolve hostname \"%s\"", bindname);
	 return STAT_RETRYLATER;
      }
      memcpy(&us.sin_addr, usaddr->h_addr_list[0], 4);
   }

   retropt_int(opts, OPT_SO_TYPE, &socktype);

   if ((result =
	_xioopen_listen(&fd->stream,
		     (struct sockaddr *)&us, sizeof(struct sockaddr_in),
		     opts, PF_INET, socktype, ipproto))
       != 0)
      return result;
   return 0;
}
#endif /* WITH_IP4 && WITH_TCP && WITH_LISTEN */

#if WITH_IP6
/* we expect the form: host:port */
int xioopen_ip6app_connect(char *a1, int rw, xiofile_t *fd,
			 unsigned groups, int socktype, int ipproto,
			 void *protname) {
   struct opt *opts = NULL;
   char *a2;
   bool needbind = false;
   struct sockaddr_in6 us;
   struct sockaddr_in6 them;
   byte2_t port;
   bool lowport = false;
   int result;

   if (a2 = strchr(a1, ',')) {
      *a2++ = '\0';
   }

   /* initialization of local vars */
   socket_in6_init(&us);
   socket_in6_init(&them);

   /* expect destination name/address */
   if ((result = parseip6addr(a1, &them, (char *)protname, 1, 1)) < 0) {
      return result;
   }

   if (parseopts(a2, groups, &opts) < 0) {
      return STAT_NORETRY;
   }
   applyopts(-1, opts, PH_INIT);

   if (retropt_2bytes(opts, OPT_SOURCEPORT, &port) >= 0) {
      us.sin6_port = htons(port);
      needbind = true;
   }

   retropt_bool(opts, OPT_LOWPORT, &lowport);

   /* 3 means: IP address AND port accepted */
   if (retropt_bind(opts, AF_INET6, (char *)protname, (struct sockaddr *)&us, 3) >= 0)
      needbind = true;

   retropt_int(opts, OPT_SO_TYPE, &socktype);

   if ((result =
	_xioopen_connect(&fd->stream,
		      needbind?(struct sockaddr *)&us:NULL, sizeof(us),
		      (struct sockaddr *)&them, sizeof(struct sockaddr_in6),
		      opts, PF_INET6, socktype, ipproto, lowport)) < 0) {
      return result;
   }
   if ((result = _xio_openlate(&fd->stream, opts)) < 0) {
      return result;
   }
   return 0;
}
#endif /* WITH_IP6 */

#endif /* WITH_TCP || WITH_UDP */
