static const char _rcsid_[]="$Id: xio-proxy.c,v 1.7 2003/02/21 08:35:02 gerhard Exp $";
/* Copyright Gerhard Rieger 2002, 2003 */
/* Published under the GNU General Public License V.2, see file COPYING */

/* this file contains the source for opening addresses of HTTP proxy CONNECT
   type */

#include "xiosysincludes.h"
#include "xioopen.h"
#include "xio-ipapp.h"

#include "xio-ascii.h"	/* for base64 encoding of authentication */
#include "xio-proxy.h"


#if WITH_PROXY

#define PROXYPORT "8080"

static int xioopen_proxy_connect(char *a1, int rw, xiofile_t *fd,
				 unsigned groups, int dummy1, int dummy2,
				 void *dummyp1);

const struct optdesc opt_proxyport = { "proxyport", NULL, OPT_PROXYPORT, GROUP_HTTP, PH_LATE, TYPE_STRING, OFUNC_SPEC };
const struct optdesc opt_ignorecr  = { "ignorecr",  NULL, OPT_IGNORECR,  GROUP_HTTP, PH_LATE, TYPE_BOOL,  OFUNC_SPEC };
const struct optdesc opt_proxy_resolve   = { "proxy-resolve",   "resolve", OPT_PROXY_RESOLVE,   GROUP_HTTP, PH_LATE, TYPE_BOOL,  OFUNC_SPEC };
const struct optdesc opt_proxy_authorization  = { "proxy-authorization",  "proxyauth", OPT_PROXY_AUTHORIZATION,  GROUP_HTTP, PH_LATE, TYPE_STRING,  OFUNC_SPEC };

const struct addrdesc addr_proxy_connect = { "proxy", 3, xioopen_proxy_connect, GROUP_FD|GROUP_SOCKET|GROUP_SOCK_IP4|GROUP_IP_TCP|GROUP_HTTP, 0, 0, NULL HELP(":<proxy-server>:<host>:<port>") };


#define CONNLEN 40	/* "CONNECT 123.156.189.123:65432 HTTP/1.0\r\n\0" */

/* states during receiving answer */
enum {
   XIOSTATE_HTTP1,	/* 0 or more bytes of first line received, no \r */
   XIOSTATE_HTTP2,	/* first line received including \r */
   XIOSTATE_HTTP3,	/* received status and \r\n */
   XIOSTATE_HTTP4,	/* within header */
   XIOSTATE_HTTP5,	/* within header, \r */
   XIOSTATE_HTTP6,	/* received status and 1 or more headers, \r\n */
   XIOSTATE_HTTP7,	/* received status line, ev. headers, \r\n\r */
   XIOSTATE_HTTP8,	/* complete answer received */
} ;


/* get buflen bytes from proxy server;
   handles EINTR;
   when error occurs, prints error message and closes fd
*/
static ssize_t
   xioproxy_recvbytes(xiofile_t *xfd, unsigned char *buff, size_t buflen) {
   ssize_t result;
   do {
      /* we need at least 16 bytes... */
      result = Read(xfd->stream.fd, buff, buflen);
   } while (result < 0 && errno == EINTR);	/*! EAGAIN? */
   if (result < 0) {
      Error4("read(%d, %p, "F_Zu"): %s",
	     xfd->stream.fd, buff, buflen, strerror(errno));
      if (Close(xfd->stream.fd) < 0) {
	 Warn2("close(%d): %s", xfd->stream.fd, strerror(errno));
      }
      return -1;
   }
   if (result == 0) {
      Error("proxy_connect: connection closed by proxy");
   }
   return result;
}


#define BUFLEN 2048

static int xioopen_proxy_connect(char *a1, int rw, xiofile_t *fd,
				 unsigned groups, int dummy1, int dummy2,
				 void *dummyp1) {
      /* we expect the form: host:host:port */
   char *a2, *a3, *a4;
   struct opt *opts = NULL;
   bool ignorecr = false;
   bool doresolve = false;
   /*struct sockaddr_in us;*/
   struct hostent *host;
   unsigned char buff[BUFLEN+1];
   size_t offset;
   char *proxyport = NULL;
   char *targetname;
   byte2_t targetport = 0;
   int socktype = SOCK_STREAM;
   char *authstring = NULL;
   unsigned char request[CONNLEN];
   unsigned char textbuff[256];
   unsigned char *eol = buff;
   int state;
   bool hstat = false;	/* HTTP status from proxy */
   int result;
   ssize_t sresult;

   a2 = strchr(a1, ':');
   if (!a2) {
      Error1("xioopen_proxy_connect(\"%s\",,): missing target address", a1);
      return STAT_NORETRY;
   }
   *a2 = '\0', ++a2;

   if ((a3 = strchr(a2, ':')) == NULL) {
      Error2("xioopen_proxy_connect(\"%s:%s\",,): missing target port", a1, a2);
      return STAT_NORETRY;
   }
   *a3 = '\0', ++a3;

   a4 = strchr(a3, ',');
   if (a4)  *a4 = '\0', ++a4;

   targetport = htons(parseport(a3, "tcp"));

   if (parseopts(a4, groups, &opts) < 0) {
      return STAT_NORETRY;
   }
   applyopts(-1, opts, PH_INIT);

   retropt_bool(opts, OPT_IGNORECR, &ignorecr);
   retropt_bool(opts, OPT_PROXY_RESOLVE, &doresolve);
   retropt_string(opts, OPT_PROXY_AUTHORIZATION, &authstring);

   if (retropt_string(opts, OPT_PROXYPORT, &proxyport) < 0) {
      if ((proxyport = strdup(PROXYPORT)) == NULL) {
	 errno = ENOMEM;  return -1;
      }
   }

   retropt_int(opts, OPT_SO_TYPE, &socktype);

   if (doresolve) {
      host = Gethostbyname(a2);
      if (host == NULL) {
	 Warn1("cannot resolve hostname for destination \"%s\",", a2);
	 targetname = strdup(a2);
      }
#define LEN 16
      if ((targetname = Malloc(16)) == NULL) {	/* www.xxx.yyy.zzz\0 */
	 return STAT_RETRYLATER;
      }
      snprintf(targetname, LEN, "%u.%u.%u.%u",
	       (unsigned char)host->h_addr_list[0][0],
	       (unsigned char)host->h_addr_list[0][1],
	       (unsigned char)host->h_addr_list[0][2],
	       (unsigned char)host->h_addr_list[0][3]);
#undef LEN
   } else {
      targetname = strdup(a2);
   }

   Notice4("opening connection to %s:%u via proxy %s:%s",
	   targetname, targetport, a1, proxyport);

   if ((result = _xioopen_ip4app_connect(a1, proxyport, fd,
					 socktype, IPPROTO_TCP, "tcp", opts))
       < 0) {
      return result;
   }

   applyopts(fd->stream.fd, opts, PH_ALL);

   if ((result = _xio_openlate(&fd->stream, opts)) < 0)
      return result;

   /* generate proxy request header - points to final target */
   sprintf(request, "CONNECT %s:%u HTTP/1.0\r\n",
	   targetname, targetport);

   /* send proxy CONNECT request (target addr+port) */
   * xiosanitize(request, strlen(request), textbuff) = '\0';
   Info1("sending \"%s\"", textbuff);
   do {
      sresult = Write(fd->stream.fd, request, strlen(request));
   } while (sresult < 0 && errno == EINTR);
   if (sresult < 0) {
      Error4("write(%d, %p, "F_Zu"): %s",
	     fd->stream.fd, request, strlen(request), strerror(errno));
      if (Close(fd->stream.fd) < 0) {
	 Warn2("close(%d): %s", fd->stream.fd, strerror(errno));
      }
      return STAT_RETRYLATER;
   }

   if (authstring) {
      /* send proxy authentication header */
      #define XIOAUTHHEAD "Proxy-authorization: Basic "
      #define XIOAUTHLEN  27
      static const char *authhead = XIOAUTHHEAD;
      #define HEADLEN 256
      char *header, *next;

      /* ...\r\n\0 */
      if ((header = Malloc(XIOAUTHLEN+((strlen(authstring)+2)/3)*4+3))
	  == NULL) {
	 return -1;
      }
      strcpy(header, authhead);
      next = xiob64encodeline(authstring, strlen(authstring),
			      strchr(header, '\0'));
      *next = '\0';
      Info1("sending \"%s\\r\\n\"", header);
      *next++ = '\r';  *next++ = '\n'; *next++ = '\0';
      do {
	 sresult = Write(fd->stream.fd, header, strlen(header));
      } while (sresult < 0 && errno == EINTR);
      if (sresult < 0) {
	 Error4("write(%d, %p, "F_Zu"): %s",
		fd->stream.fd, header, strlen(header), strerror(errno));
	 if (Close(fd->stream.fd) < 0) {
	    Warn2("close(%d): %s", fd->stream.fd, strerror(errno));
	 }
	 return STAT_RETRYLATER;
      }

      free(header);
   }

   Info("sending \"\\r\\n\"");
   do {
      sresult = Write(fd->stream.fd, "\r\n", 2);
   } while (sresult < 0 && errno == EINTR);

   /* request is kept for later error messages */
   *strstr(request, " HTTP") = '\0';

   /* receive proxy answer; looks like "HTTP/1.0 200 .*\r\nHeaders..\r\n\r\n" */
   /* socat version 1 depends on a valid fd for data transfer; address
      therefore cannot buffer data. So, to prevent reading beyond the end of
      the answer headers, only single bytes are read. puh. */
   state = XIOSTATE_HTTP1;
   offset = 0;	/* up to where the buffer is filled (relative) */
   /*eol;*/	/* points to the first lineterm of the current line */
   do {
      sresult = xioproxy_recvbytes(fd, buff+offset, 1);
      if (sresult <= 0) {
	 return STAT_RETRYLATER;
      }
      
      switch (state) {

      case XIOSTATE_HTTP1:
	 /* 0 or more bytes of first line received, no '\r' yet */
	 if (*(buff+offset) == '\r') {
	    eol = buff+offset;
	    state = XIOSTATE_HTTP2;
	    break;
	 }
	 if (ignorecr && *(buff+offset) == '\n') {
	    eol = buff+offset;
	    state = XIOSTATE_HTTP3;
	    break;
	 }
	 break;

      case XIOSTATE_HTTP2:
	 /* first line received including '\r' */
	 if (*(buff+offset) != '\n') {
	    state = XIOSTATE_HTTP1;
	    break;
	 }
	 state = XIOSTATE_HTTP3;
	 break;

      case XIOSTATE_HTTP3:
	 /* received status (first line) and "\r\n" */
	 if (*(buff+offset) == '\r') {
	    state = XIOSTATE_HTTP7;
	    break;
	 }
	 if (ignorecr && *(buff+offset) == '\n') {
	    state = XIOSTATE_HTTP8;
	    break;
	 }
	 state = XIOSTATE_HTTP4;
	 break;
	 
      case XIOSTATE_HTTP4:
	 /* within header */
	 if (*(buff+offset) == '\r') {
	    eol = buff+offset;
	    state = XIOSTATE_HTTP5;
	    break;
	 }
	 if (ignorecr && *(buff+offset) == '\n') {
	    eol = buff+offset;
	    state = XIOSTATE_HTTP6;
	    break;
	 }
	 break;

      case XIOSTATE_HTTP5:
	 /* within header, '\r' received */
	 if (*(buff+offset) != '\n') {
	    state = XIOSTATE_HTTP4;
	    break;
	 }
	 state = XIOSTATE_HTTP6;
	 break;

      case XIOSTATE_HTTP6:
	 /* received status (first line) and 1 or more headers, "\r\n" */
	 if (*(buff+offset) == '\r') {
	    state = XIOSTATE_HTTP7;
	    break;
	 }
	 if (ignorecr && *(buff+offset) == '\n') {
	    state = XIOSTATE_HTTP8;
	    break;
	 }
	 state = XIOSTATE_HTTP4;
	 break;

      case XIOSTATE_HTTP7:
	 /* received status (first line), 0 or more headers, "\r\n\r" */
	 if (*(buff+offset) == '\n') {
	    state = XIOSTATE_HTTP8;
	    break;
	 }
	 if (*(buff+offset) == '\r') {
	    if (ignorecr) {
	       break;	/* ignore it, keep waiting for '\n' */
	    } else {
	       state = XIOSTATE_HTTP5;
	    }
	    break;
	 }
	 state = XIOSTATE_HTTP4;
	 break;

      }
      ++offset;

      /* end of status line reached */
      if (state == XIOSTATE_HTTP3) {
	 /* set a terminating null - on or after CRLF? */
	 *(buff+offset) = '\0';

	 * xiosanitize(buff, Min(offset, (sizeof(textbuff)-1)>>1), textbuff)
	    = '\0';
	 Info1("proxy_connect: received answer \"%s\"", textbuff);
	 *eol = '\0';
	 * xiosanitize(buff, Min(strlen(buff), (sizeof(textbuff)-1)>>1),
		       textbuff) = '\0';
	 if (strncmp(buff, "HTTP/1.0 ", 9) &&
	     strncmp(buff, "HTTP/1.1 ", 9)) {
	    /* invalid answer */
	    Error1("proxy: invalid answer \"%s\"", textbuff);
	    if (Close(fd->stream.fd) < 0) {
	       Warn2("close(%d): %s", fd->stream.fd, strerror(errno));
	    }
	    return STAT_NORETRY;
	 }

	 /* HTTP answer */
	 if (strncmp(buff+9, "200", 3)) {
	    /*! not ok */
	    /* CERN:
	       "HTTP/1.0 200 Connection established"
	       "HTTP/1.0 400 Invalid request "CONNECT 10.244.9.3:8080 HTTP/1.0" (unknown method)"
	       "HTTP/1.0 403 Forbidden - by rule"
	       "HTTP/1.0 407 Proxy Authentication Required"
	       Proxy-Authenticate: Basic realm="Squid proxy-caching web server"
>  50 72 6f 78 79 2d 61 75 74 68 6f 72 69 7a 61 74  Proxy-authorizat
>  69 6f 6e 3a 20 42 61 73 69 63 20 61 57 4e 6f 63  ion: Basic aWNoc
>  32 56 73 59 6e 4e 30 4f 6e 4e 30 63 6d 56 75 5a  2VsYnN0OnN0cmVuZ
>  32 64 6c 61 47 56 70 62 51 3d 3d 0d 0a           2dlaGVpbQ==..
                b64encode("username:password")
	       "HTTP/1.0 500 Can't connect to host"
	    */
	    /* Squid:
	       "HTTP/1.0 400 Bad Request"
	       "HTTP/1.0 403 Forbidden"
	       "HTTP/1.0 503 Service Unavailable"
	       interesting header: "X-Squid-Error: ERR_CONNECT_FAIL 111" */
	    /* Apache:
	       "HTTP/1.0 400 Bad Request"
	       "HTTP/1.1 405 Method Not Allowed"
	    */
	    /* WTE:
	       "HTTP/1.1 200 Connection established"
	       "HTTP/1.1 404 Host not found or not responding, errno:  79"
	       "HTTP/1.1 404 Host not found or not responding, errno:  32"
	       "HTTP/1.1 404 Host not found or not responding, errno:  13"
	    */
	    /* IIS:
	       "HTTP/1.1 404 Object Not Found"
	    */
	    Error2("%s: %s", request, textbuff+13);
	    hstat = false;
	 } else {

	    /*! ok!! */
	    /* "HTTP/1.0 200 Connection established" */
	    /*Info1("proxy: \"%s\"", textbuff+13);*/
	    hstat = true;
	 }
	 offset = 0;

      } else if (state == XIOSTATE_HTTP6) {
      /* end of a header line reached */
	 char *endp;

	 /* set a terminating null */
	 *(buff+offset) = '\0';

	 endp =
	    xiosanitize(buff, Min(offset, (sizeof(textbuff)-1)>>1),
			textbuff);
	 *endp = '\0';
	 Info1("proxy_connect: received header \"%s\"", textbuff);
	 offset = 0;
      }

   } while (state != XIOSTATE_HTTP8 && offset < BUFLEN);

   if (offset >= BUFLEN) {
      Error1("proxy answer exceeds "F_Zu" bytes, aborting", BUFLEN);
      if (Close(fd->stream.fd) < 0) {
	 Warn2("close(%d): %s", fd->stream.fd, strerror(errno));
      }
      return STAT_NORETRY;
   }

   if (!hstat) {
      if (Close(fd->stream.fd) < 0) {
	 Warn2("close(%d): %s", fd->stream.fd, strerror(errno));
      }
      return STAT_RETRYLATER;
   }

   Notice4("successfully connected to %s:%s via proxy %s:%s", a2, a3,
	   a1, proxyport);

   return 0;
}

#endif /* WITH_PROXY */

