static const char _rcsid_[]="$Id: xio-socks.c,v 1.14 2002/10/19 20:36:51 gerhard Exp $";
/* Copyright Gerhard Rieger 2001, 2002 */
/* Published under the GNU General Public License V.2, see file COPYING */

/* this file contains the source for opening addresses of socks4 type */

#include "xiosysincludes.h"
#include "xioopen.h"
#include "xio-ipapp.h"

#include "xio-socks.h"


#if WITH_SOCKS4 || WITH_SOCKS4A
enum {
  SOCKS_CD_GRANTED = 90,
  SOCKS_CD_FAILED,
  SOCKS_CD_NOIDENT,
  SOCKS_CD_IDENTFAILED
} ;

#define SOCKSPORT "1080"

static int xioopen_socks4_connect(char *a1, int rw, xiofile_t *fd,
				unsigned groups, int dummy1, int dummy2,
				void *dummyp1);

const struct optdesc opt_socksport = { "socksport", NULL, OPT_SOCKSPORT, GROUP_IP_SOCKS4, PH_LATE, TYPE_STRING, OFUNC_SPEC };
const struct optdesc opt_socksuser = { "socksuser", NULL, OPT_SOCKSUSER, GROUP_IP_SOCKS4, PH_LATE, TYPE_NAME, OFUNC_SPEC };

const struct addrdesc addr_socks4_connect = { "socks4", 3, xioopen_socks4_connect, GROUP_FD|GROUP_SOCKET|GROUP_SOCK_IP4|GROUP_IP_TCP|GROUP_IP_SOCKS4, 0, 0, NULL HELP(":<socks-server>:<host>:<port>") };

const struct addrdesc addr_socks4a_connect = { "socks4a", 3, xioopen_socks4_connect, GROUP_FD|GROUP_SOCKET|GROUP_SOCK_IP4|GROUP_IP_TCP|GROUP_IP_SOCKS4, 1, 0, NULL HELP(":<socks-server>:<host>:<port>") };


static int xioopen_socks4_connect(char *a1, int rw, xiofile_t *fd,
				unsigned groups, int dummy1, int dummy2,
				void *dummyp1) {
      /* we expect the form: host:host:port */
   char *a2, *a3, *a4;
   struct opt *opts = NULL;
   struct sockaddr_in us;
   socklen_t uslen;
   struct servent *se;
   struct hostent *host;
   unsigned char buff[8+256];
   struct socks4 {
      unsigned char version;
      unsigned char action;
      unsigned short port;
      unsigned long dest;
      /* const char userid[]; */
   } *sockhead = (struct socks4 *)buff;
   char *userid;
   size_t buflen;
   ssize_t bytes;
   char *socksport = NULL;
   int socktype = SOCK_STREAM;
   unsigned char infobuff[256];
   int result;

   socket_in_init(&us);

   a2 = strchr(a1, ':');
   if (!a2) {
      Error1("xioopen_socks4_connect(\"%s\",,): missing target address", a1);
      return STAT_NORETRY;
   }
   *a2 = '\0', ++a2;

   if ((a3 = strchr(a2, ':')) == NULL) {
      Error2("xioopen_socks4_connect(\"%s:%s\",,): missing target port", a1, a2);
      return STAT_NORETRY;
   }
   *a3 = '\0', ++a3;

   host = Gethostbyname(a2);
   if (host == NULL) {
#if !WITH_SOCKS4A
      Error1("cannot resolve hostname for destination \"%s\",", a2);
      return STAT_RETRYLATER;
#else /* WITH_SOCKS4A */
      Info1("cannot resolve hostname for destination \"%s\",", a2);
#endif /* WITH_SOCKS4A */
   }

   a4 = strchr(a3, ',');
   if (a4)  *a4 = '\0', ++a4;

   /* generate socks header - points to final target */
   sockhead->version = 4;
   sockhead->action  = 1;
   sockhead->port    = parseport(a3, "tcp");

#if !WITH_SOCKS4A
   memcpy(&sockhead->dest, host->h_addr_list[0], 4);
#else /* WITH_SOCKS4A */
   if (host != NULL) {
      memcpy(&sockhead->dest, host->h_addr_list[0], 4);
   } else {
      sockhead->dest = htonl(0x00000001);	/* three bytes zero */
   }
#endif /* WITH_SOCKS4A */

   if (parseopts(a4, groups, &opts) < 0) {
      return STAT_NORETRY;
   }

   if (retropt_string(opts, OPT_SOCKSPORT, &socksport) < 0) {
      if ((se = getservbyname("socks", "tcp")) != NULL) {
	 Debug1("\"socks/tcp\" resolves to %u", ntohs(se->s_port));
	 if ((socksport = Malloc(6)) == NULL) {
	    return -1;
	 }
	 sprintf(socksport, "%u", ntohs(se->s_port));
      } else {
	 Debug1("cannot resolve service \"socks/tcp\", using %s", SOCKSPORT);
	 if ((socksport = strdup(SOCKSPORT)) == NULL) {
	    errno = ENOMEM;  return -1;
	 }
      }
   }

   if (retropt_string(opts, OPT_SOCKSUSER, &userid) < 0) {
      if ((userid = getenv("LOGNAME")) == NULL) {
	 if ((userid = getenv("USER")) == NULL) {
	    userid = "anonymous";
	 }
      }
   }
   strncpy(buff+sizeof(struct socks4), userid, 256);
   buflen = sizeof(struct socks4)+strlen(userid)+1;
#if WITH_SOCKS4A
   /* this has not been tested */
   if (host == NULL) {
      strncpy(buff+buflen, a2, 256+8-buflen);
      buflen += strlen(a2) + 1;
   }
#endif /* WITH_SOCKS4A */

   retropt_int(opts, OPT_SO_TYPE, &socktype);

   Notice8("Opening connection to %u.%u.%u.%u:%u via socks4 server %s:%s as user \"%s\"",
	   ((unsigned char *)&sockhead->dest)[0],
	   ((unsigned char *)&sockhead->dest)[1],
	   ((unsigned char *)&sockhead->dest)[2],
	   ((unsigned char *)&sockhead->dest)[3],
	   ntohs(sockhead->port),
	   a1, socksport, userid);

   result =
      _xioopen_ip4app_connect(a1, socksport, fd, socktype, IPPROTO_TCP, "tcp",
			      opts);

   applyopts(fd->stream.fd, opts, PH_ALL);

   if ((result = _xio_openlate(&fd->stream, opts)) < 0)
      return result;

   /* send socks header (target addr+port, +auth) */
   Info("sending socks request");
   do {
      result = Write(fd->stream.fd, buff, buflen);
   } while (result < 0 && errno == EINTR);
   if (result < 0) {
      Error4("write(%d, %p, "F_Zu"): %s",
	     fd->stream.fd, buff, buflen, strerror(errno));
      if (Close(fd->stream.fd) < 0) {
	 Warn2("close(%d): %s", fd->stream.fd, strerror(errno));
      }
      return STAT_RETRYLATER;
   }

   bytes = 0;
   Info("waiting for socks reply");
   while (true) {
      /* receive socks answer */
      do {
	 result = Read(fd->stream.fd, buff+bytes, sizeof(struct socks4)-bytes);
      } while (result < 0 && errno == EINTR);
      if (result < 0) {
	 Error4("read(%d, %p, "F_Zu"): %s",
		fd->stream.fd, buff+bytes, sizeof(struct socks4)-bytes,
		strerror(errno));
	 if (Close(fd->stream.fd) < 0) {
	    Warn2("close(%d): %s", fd->stream.fd, strerror(errno));
	 }
	 return STAT_RETRYLATER;
      }
      if (result == 0) {
	 Error("read(): EOF during read of socks reply");
	 return STAT_RETRYLATER;
      }
      bytes += result;
      if (bytes == sizeof(struct socks4)) {
	 Debug1("received all "F_Zd" bytes", bytes);
	 break;
      }
      Debug2("received "F_Zd" bytes, waiting for "F_Zu" more bytes",
	     result, sizeof(struct socks4)-bytes);
   }
   Info7("socks answer: {%u, %u, %u, %u.%u.%u.%u}",
	 sockhead->version, sockhead->action, ntohs(sockhead->port),
	 ((uint8_t *)&sockhead->dest)[0],
	 ((uint8_t *)&sockhead->dest)[1],
	 ((uint8_t *)&sockhead->dest)[2],
	 ((uint8_t *)&sockhead->dest)[3]);
   if (sockhead->version != 0) {
      Warn1("socks: reply code version is not 0 (%d)",
	    sockhead->version);
   }
   switch (sockhead->action) {
   case SOCKS_CD_GRANTED:
      /* Notice("socks: connect request succeeded"); */
      uslen = sizeof(us);
      if (Getsockname(fd->stream.fd, (struct sockaddr *)&us, &uslen) < 0) {
	 Error4("getsockname(%d, %p, {%d}): %s",
		fd->stream.fd, &us, uslen, strerror(errno));
      }
      Notice1("successfully connected from %s via socks",
	      sockaddr_info((struct sockaddr *)&us, infobuff, sizeof(infobuff)));
      break;
   case SOCKS_CD_FAILED:
      Error("socks: connect request rejected or failed");
      return STAT_RETRYLATER;
   case SOCKS_CD_NOIDENT:
      Error("socks: ident refused by client");
      return STAT_RETRYLATER;
   case SOCKS_CD_IDENTFAILED:
      Error("socks: ident failed");
      return STAT_RETRYLATER;
   default:
      Error1("socks: undefined status %u", sockhead->action);
   }
   return 0;
}
#endif /* WITH_SOCKS4 || WITH_SOCKS4A */

