static const char _rcsid_[]="$Id: xioopts.c,v 1.47 2003/02/23 08:33:41 gerhard Exp $";
/* Copyright Gerhard Rieger 2001-2003 */
/* Published under the GNU General Public License V.2, see file COPYING */

/* this file contains the source for address options handling */

#include "xiosysincludes.h"
#include "xioopen.h"

#include "xiomodes.h"


bool xioopts_ignoregroups;

#define IF_ANY(a,b) {a,b},

#if WITH_NAMED
#  define IF_NAMED(a,b) {a,b},
#else
#  define IF_NAMED(a,b) 
#endif

#if WITH_PIPE || WITH_GOPEN
#  define IF_OPEN(a,b) {a,b},
#else
#  define IF_OPEN(a,b) 
#endif

#if WITH_TERMIOS
#  define IF_TERMIOS(a,b) {a,b},
#else
#  define IF_TERMIOS(a,b) 
#endif

#if WITH_EXEC
#  define IF_EXEC(a,b) {a,b},
#else
#  define IF_EXEC(a,b) 
#endif

#if WITH_SOCKET
#  define IF_SOCKET(a,b) {a,b},
#else
#  define IF_SOCKET(a,b) 
#endif

#if WITH_LISTEN
#  define IF_LISTEN(a,b) {a,b},
#else
#  define IF_LISTEN(a,b) 
#endif

#if (WITH_UDP || WITH_TCP) && WITH_LISTEN
#  define IF_RANGE(a,b) {a,b},
#else
#  define IF_RANGE(a,b)
#endif

#if WITH_IP4 || WITH_IP6
#  define IF_IP(a,b) {a,b},
#else
#  define IF_IP(a,b) 
#endif

#if WITH_TCP|WITH_UDP
#  define IF_IPAPP(a,b) {a,b},
#else
#  define IF_IPAPP(a,b) 
#endif

#if WITH_TCP
#  define IF_TCP(a,b) {a,b},
#else
#  define IF_TCP(a,b) 
#endif

#if WITH_SOCKS4
#  define IF_SOCKS4(a,b) {a,b},
#else
#  define IF_SOCKS4(a,b) 
#endif

#if WITH_PROXY
#  define IF_PROXY(a,b) {a,b},
#else
#  define IF_PROXY(a,b) 
#endif

#if WITH_READLINE
#  define IF_READLINE(a,b) {a,b},
#else
#  define IF_READLINE(a,b)
#endif

#if WITH_PTY
#  define IF_PTY(a,b) {a,b},
#else
#  define IF_PTY(a,b)
#endif

#if WITH_OPENSSL
#  define IF_OPENSSL(a,b) {a,b},
#else
#  define IF_OPENSSL(a,b)
#endif

#if WITH_RETRY
#  define IF_RETRY(a,b) {a,b},
#else
#  define IF_RETRY(a,b) 
#endif


/* address options - keep this array strictly alphabetically sorted for
   binary search! */
/* NULL terminated */
const struct optname optionnames[] = {
#ifdef SO_ACCEPTCONN /* AIX433 */
	IF_SOCKET ("acceptconn",	&opt_so_acceptconn)
#endif /* SO_ACCEPTCONN */
	IF_ANY    ("append",	&opt_append)
#ifdef O_ASYNC
	IF_ANY    ("async",	&opt_async)
#endif
#ifdef SO_ATTACH_FILTER
	IF_SOCKET ("attach-filter",	&opt_so_attach_filter)
	IF_SOCKET ("attachfilter",	&opt_so_attach_filter)
#endif
#ifdef SO_AUDIT	/* AIX 4.3.3 */
	IF_SOCKET ("audit",	&opt_so_audit)
#endif /* SO_AUDIT */
#ifdef CBAUD
	IF_TERMIOS("b0",	&opt_b0)
#ifdef B1000000
	IF_TERMIOS("b1000000",	&opt_b1000000)
#endif
	IF_TERMIOS("b110",	&opt_b110)
#ifdef B115200
	IF_TERMIOS("b115200",	&opt_b115200)
#endif
#ifdef B1152000
	IF_TERMIOS("b1152000",	&opt_b1152000)
#endif
	IF_TERMIOS("b1200",	&opt_b1200)
	IF_TERMIOS("b134",	&opt_b134)
	IF_TERMIOS("b150",	&opt_b150)
#ifdef B1500000
	IF_TERMIOS("b1500000",	&opt_b1500000)
#endif
	IF_TERMIOS("b1800",	&opt_b1800)
	IF_TERMIOS("b19200",	&opt_b19200)
	IF_TERMIOS("b200",	&opt_b200)
#ifdef B2000000
	IF_TERMIOS("b2000000",	&opt_b2000000)
#endif
#ifdef B230400
	IF_TERMIOS("b230400",	&opt_b230400)
#endif
	IF_TERMIOS("b2400",	&opt_b2400)
#ifdef B2500000
	IF_TERMIOS("b2500000",	&opt_b2500000)
#endif
	IF_TERMIOS("b300",	&opt_b300)
#ifdef B3000000
	IF_TERMIOS("b3000000",	&opt_b3000000)
#endif
#ifdef B3500000
	IF_TERMIOS("b3500000",	&opt_b3500000)
#endif
	IF_TERMIOS("b38400",	&opt_b38400)
#ifdef B4000000
	IF_TERMIOS("b4000000",	&opt_b4000000)
#endif
#ifdef B460800
	IF_TERMIOS("b460800",	&opt_b460800)
#endif
	IF_TERMIOS("b4800",	&opt_b4800)
	IF_TERMIOS("b50",	&opt_b50)
#ifdef B500000
	IF_TERMIOS("b500000",	&opt_b500000)
#endif
#ifdef B57600
	IF_TERMIOS("b57600",	&opt_b57600)
#endif
#ifdef B576000
	IF_TERMIOS("b576000",	&opt_b576000)
#endif
	IF_TERMIOS("b600",	&opt_b600)
	IF_TERMIOS("b75",	&opt_b75)
#ifdef B921600
	IF_TERMIOS("b921600",	&opt_b921600)
#endif
	IF_TERMIOS("b9600",	&opt_b9600)
#endif /* defined(CBAUD) */
	IF_LISTEN ("backlog",	&opt_backlog)
	IF_SOCKET ("bind",	&opt_bind)
#ifdef SO_BINDTODEVICE
	IF_SOCKET ("bindtodevice",	&opt_so_bindtodevice)
#endif
	IF_TERMIOS("brkint",	&opt_brkint)
	IF_SOCKET ("broadcast",	&opt_so_broadcast)
#ifdef BSDLY
#  ifdef BS0
	IF_TERMIOS("bs0",	&opt_bs0)
#  endif
#  ifdef BS1
	IF_TERMIOS("bs1",	&opt_bs1)
#  endif
#endif
#ifdef SO_BSDCOMPAT
	IF_SOCKET ("bsdcompat",	&opt_so_bsdcompat)
#endif
#ifdef BSDLY
	IF_TERMIOS("bsdly",	&opt_bsdly)
#endif
	IF_ANY    ("chroot",	&opt_chroot)
	IF_ANY    ("chroot-early",	&opt_chroot_early)
	/*IF_TERMIOS("cibaud",	&opt_cibaud)*/
	IF_OPENSSL("cipherlist",	&opt_openssl_cipherlist)
	IF_OPENSSL("ciphers",	&opt_openssl_cipherlist)
#ifdef SO_CKSUMRECV
	IF_SOCKET ("cksumrecv",	&opt_so_cksumrecv)
#endif /* SO_CKSUMRECV */
	/*IF_NAMED  ("cleanup",	&opt_cleanup)*/
	IF_TERMIOS("clocal",	&opt_clocal)
	IF_ANY    ("cloexec",	&opt_cloexec)
#ifdef TCP_CORK
	IF_TCP    ("cork",	&opt_tcp_cork)
#endif
	IF_ANY    ("cr",        &opt_cr)
#ifdef CRDLY
#  ifdef CR0
	IF_TERMIOS("cr0",	&opt_cr0)
#  endif
#  ifdef CR1
	IF_TERMIOS("cr1",	&opt_cr1)
#  endif
#  ifdef CR2
	IF_TERMIOS("cr2",	&opt_cr2)
#  endif
#  ifdef CR3
	IF_TERMIOS("cr3",	&opt_cr3)
#  endif
	IF_TERMIOS("crdly",	&opt_crdly)
#endif /* defined(CRDLY) */
	IF_TERMIOS("cread",	&opt_cread)
	IF_OPEN   ("creat",	&opt_o_create)
	IF_OPEN   ("create",	&opt_o_create)
	IF_ANY    ("crlf",      &opt_crnl)
	IF_ANY    ("crnl",      &opt_crnl)
	IF_TERMIOS("crterase",	&opt_echoe)
	IF_TERMIOS("crtkill",	&opt_echoke)
#ifdef CRTSCTS
	IF_TERMIOS("crtscts",	&opt_crtscts)
#endif
	IF_TERMIOS("cs5",	&opt_cs5)
	IF_TERMIOS("cs6",	&opt_cs6)
	IF_TERMIOS("cs7",	&opt_cs7)
	IF_TERMIOS("cs8",	&opt_cs8)
	IF_TERMIOS("csize",	&opt_csize)
	IF_TERMIOS("cstopb",	&opt_cstopb)
	IF_TERMIOS("ctlecho",	&opt_echoctl)
	IF_TERMIOS("ctty",	&opt_tiocsctty)
	IF_EXEC   ("dash",	&opt_dash)
	IF_SOCKET ("debug",	&opt_so_debug)
#ifdef O_DEFER
	IF_OPEN   ("defer",	&opt_o_defer)
#endif
#ifdef TCP_DEFER_ACCEPT	/* Linux 2.4.0 */
	IF_TCP    ("defer-accept",	&opt_tcp_defer_accept)
#endif
#ifdef O_DELAY
	IF_OPEN   ("delay",	&opt_o_delay)
#endif
	IF_NAMED  ("delete",	&opt_unlink)
#ifdef SO_DETACH_FILTER
	IF_SOCKET ("detach-filter", &opt_so_detach_filter)
	IF_SOCKET ("detachfilter",  &opt_so_detach_filter)
#endif
#ifdef SO_DGRAM_ERRIND
	IF_SOCKET ("dgram-errind",	&opt_so_dgram_errind)
	IF_SOCKET ("dgramerrind",	&opt_so_dgram_errind)
#endif
#ifdef O_DIRECT
	IF_OPEN   ("direct",	&opt_o_direct)
#endif
#ifdef O_DIRECTORY
	IF_OPEN   ("directory",	&opt_o_directory)
#endif
#ifdef VDISCARD
	IF_TERMIOS("discard",	&opt_vdiscard)
#endif
#ifdef SO_DONTLINGER
	IF_SOCKET ("dontlinger",	&opt_so_dontlinger)
#endif
	IF_SOCKET ("dontroute",	&opt_so_dontroute)
#ifdef O_DSYNC
	IF_OPEN   ("dsync",	&opt_o_dsync)
#endif
	IF_TERMIOS("echo",	&opt_echo)
	IF_TERMIOS("echoctl",	&opt_echoctl)
	IF_TERMIOS("echoe",	&opt_echoe)
	IF_TERMIOS("echok",	&opt_echok)
	IF_TERMIOS("echoke",	&opt_echoke)
	IF_TERMIOS("echonl",	&opt_echonl)
#ifdef ECHOPRT
	IF_TERMIOS("echoprt",	&opt_echoprt)
#endif
	IF_TERMIOS("eof",	&opt_veof)
	IF_TERMIOS("eol",	&opt_veol)
	IF_TERMIOS("eol2",	&opt_veol2)
	IF_TERMIOS("erase",	&opt_verase)
	IF_SOCKET ("error",	&opt_so_error)
	IF_OPEN   ("excl",	&opt_o_excl)
	IF_ANY 	  ("f-setlk",	&opt_f_setlk)
	IF_ANY 	  ("f-setlkw",	&opt_f_setlkw)
	IF_EXEC   ("fdin",	&opt_fdin)
	IF_EXEC   ("fdout",	&opt_fdout)
#ifdef FFDLY
#  ifdef FF0
	IF_TERMIOS("ff0",	&opt_ff0)
#  endif
#  ifdef FF1
	IF_TERMIOS("ff1",	&opt_ff1)
#  endif
	IF_TERMIOS("ffdly",	&opt_ffdly)
#endif
#ifdef FIOSETOWN
	IF_SOCKET ("fiosetown",	&opt_fiosetown)
#endif
#if HAVE_FLOCK
	IF_ANY    ("flock",	&opt_flock_ex)
	IF_ANY    ("flock-ex",	&opt_flock_ex)
	IF_ANY    ("flock-ex-nb",	&opt_flock_ex_nb)
	IF_ANY    ("flock-nb",	&opt_flock_ex_nb)
	IF_ANY    ("flock-sh",	&opt_flock_sh)
	IF_ANY    ("flock-sh-nb",	&opt_flock_sh_nb)
#endif
	IF_TERMIOS("flusho",	&opt_flusho)
	/*0 IF_NAMED  ("force",	&opt_force)*/
	IF_RETRY  ("forever",	&opt_forever)
	IF_LISTEN ("fork",	&opt_fork)
#ifdef IP_FREEBIND
	IF_IP     ("freebind",	&opt_ip_freebind)
#endif
	IF_ANY    ("ftruncate",	&opt_ftruncate)
	IF_ANY    ("gid",	&opt_group)
	IF_NAMED  ("gid-e",	&opt_group_early)
	IF_ANY    ("gid-l",	&opt_group_late)
	IF_ANY    ("group",	&opt_group)
	IF_NAMED  ("group-early",	&opt_group_early)
	IF_ANY    ("group-late",	&opt_group_late)
#ifdef IP_HDRINCL
	IF_IP     ("hdrincl",	&opt_ip_hdrincl)
#endif
	IF_READLINE("history",	&opt_history_file)
	IF_READLINE("history-file",	&opt_history_file)
	IF_TERMIOS("hup",	&opt_hupcl)
	IF_TERMIOS("hupcl",	&opt_hupcl)
	IF_TERMIOS("icanon",	&opt_icanon)
	IF_TERMIOS("icrnl",	&opt_icrnl)
	IF_TERMIOS("iexten",	&opt_iexten)
#ifdef SO_BINDTODEVICE
	IF_SOCKET ("if",	&opt_so_bindtodevice)
#endif
	IF_TERMIOS("ignbrk",	&opt_ignbrk)
	IF_TERMIOS("igncr",	&opt_igncr)
  /* you might need to terminate socat manually if you use this option: */
	IF_PROXY  ("ignorecr",	&opt_ignorecr)
	IF_ANY    ("ignoreeof",	&opt_ignoreeof)
	IF_ANY    ("ignoreof",	&opt_ignoreeof)
	IF_TERMIOS("ignpar",	&opt_ignpar)
	IF_TERMIOS("imaxbel",	&opt_imaxbel)
#ifdef TCP_INFO	/* Linux 2.4.0 */
	IF_TCP    ("info",	&opt_tcp_info)
#endif
	IF_TERMIOS("inlcr",	&opt_inlcr)
	IF_TERMIOS("inpck",	&opt_inpck)
#ifdef SO_BINDTODEVICE
	IF_SOCKET ("interface",	&opt_so_bindtodevice)
#endif
	IF_RETRY  ("intervall",	&opt_intervall)
	IF_TERMIOS("intr",	&opt_vintr)
#ifdef IP_FREEBIND
	IF_IP     ("ip-freebind",	&opt_ip_freebind)
#endif
#ifdef IP_HDRINCL
	IF_IP     ("ip-hdrincl",	&opt_ip_hdrincl)
#endif
#ifdef IP_MTU
	IF_IP     ("ip-mtu",	&opt_ip_mtu)
#endif
#ifdef IP_MTU_DISCOVER
	IF_IP     ("ip-mtu-discover",	&opt_ip_mtu_discover)
#endif
	IF_IP     ("ip-multicast-loop",	&opt_ip_multicast_loop)
	IF_IP     ("ip-multicast-ttl",	&opt_ip_multicast_ttl)
#ifdef IP_OPTIONS
	IF_IP     ("ip-options",	&opt_ip_options)
#endif
#ifdef IP_PKTINFO
	IF_IP     ("ip-pktinfo",	&opt_ip_pktinfo)
#endif
#ifdef IP_PKTOPTIONS
	IF_IP     ("ip-pktoptions",	&opt_ip_pktoptions)
#endif
#ifdef IP_RECVERR
	IF_IP     ("ip-recverr",	&opt_ip_recverr)
#endif
#ifdef IP_RECVOPTS
	IF_IP     ("ip-recvopts",	&opt_ip_recvopts)
#endif
#ifdef IP_RECVTOS
	IF_IP     ("ip-recvtos",	&opt_ip_recvtos)
#endif
#ifdef IP_RECVTTL
	IF_IP     ("ip-recvttl",	&opt_ip_recvttl)
#endif
#ifdef IP_RETOPTS
	IF_IP     ("ip-retopts",	&opt_ip_retopts)
#endif
#ifdef IP_ROUTER_ALERT
	IF_IP     ("ip-router-alert",	&opt_ip_router_alert)
#endif
	IF_IP     ("ip-tos",	&opt_ip_tos)
	IF_IP     ("ip-ttl",	&opt_ip_ttl)
#ifdef IP_FREEBIND
	IF_IP     ("ipfreebind",	&opt_ip_freebind)
#endif
#ifdef IP_HDRINCL
	IF_IP     ("iphdrincl",	&opt_ip_hdrincl)
#endif
#ifdef IP_MTU
	IF_IP     ("ipmtu",	&opt_ip_mtu)
#endif
#ifdef IP_MTU_DISCOVER
	IF_IP     ("ipmtudiscover",	&opt_ip_mtu_discover)
#endif
	IF_IP     ("ipmulticastloop",	&opt_ip_multicast_loop)
	IF_IP     ("ipmulticastttl",	&opt_ip_multicast_ttl)
#ifdef IP_OPTIONS
	IF_IP     ("ipoptions",	&opt_ip_options)
#endif
#ifdef IP_PKTINFO
	IF_IP     ("ippktinfo",	&opt_ip_pktinfo)
#endif
#ifdef IP_PKTOPTIONS
	IF_IP     ("ippktoptions",	&opt_ip_pktoptions)
#endif
#ifdef IP_RECVERR
	IF_IP     ("iprecverr",	&opt_ip_recverr)
#endif
#ifdef IP_RECVOPTS
	IF_IP     ("iprecvopts",	&opt_ip_recvopts)
#endif
#ifdef IP_RECVTOS
	IF_IP     ("iprecvtos",	&opt_ip_recvtos)
#endif
#ifdef IP_RECVTTL
	IF_IP     ("iprecvttl",	&opt_ip_recvttl)
#endif
#ifdef IP_RETOPTS
	IF_IP     ("ipretopts",	&opt_ip_retopts)
#endif
#ifdef IP_ROUTER_ALERT
	IF_IP     ("iprouteralert",	&opt_ip_router_alert)
#endif
	IF_IP     ("iptos",	&opt_ip_tos)
	IF_IP     ("ipttl",	&opt_ip_ttl)
	IF_TERMIOS("isig",	&opt_isig)
#ifdef HAVE_TERMIOS_ISPEED
	IF_TERMIOS("ispeed",	&opt_ispeed)
#endif
	IF_TERMIOS("istrip",	&opt_istrip)
#ifdef IUCLC
	IF_TERMIOS("iuclc",	&opt_iuclc)
#endif
	IF_TERMIOS("ixany",	&opt_ixany)
	IF_TERMIOS("ixoff",	&opt_ixoff)
	IF_TERMIOS("ixon",	&opt_ixon)
	IF_SOCKET ("keepalive",	&opt_so_keepalive)
#ifdef TCP_KEEPCNT	/* Linux 2.4.0 */
	IF_TCP    ("keepcnt",	&opt_tcp_keepcnt)
#endif
#ifdef TCP_KEEPIDLE	/* Linux 2.4.0 */
	IF_TCP    ("keepidle",	&opt_tcp_keepidle)
#endif
#ifdef TCP_KEEPINTVL	/* Linux 2.4.0 */
	IF_TCP    ("keepintvl",	&opt_tcp_keepintvl)
#endif
#ifdef SO_KERNACCEPT	/* AIX 4.3.3 */
	IF_SOCKET ("kernaccept",	&opt_so_kernaccept)
#endif /* SO_KERNACCEPT */
	IF_TERMIOS("kill",	&opt_vkill)
#ifdef O_LARGEFILE
	IF_OPEN   ("largefile",	&opt_o_largefile)
#endif
#if WITH_LIBWRAP
	IF_IPAPP  ("libwrap",		&opt_tcpwrappers)
#endif
	IF_SOCKET ("linger",	&opt_so_linger)
#ifdef TCP_LINGER2	/* Linux 2.4.0 */
	IF_TCP    ("linger2",	&opt_tcp_linger2)
#endif
	IF_PTY    ("link",	&opt_symbolic_link)
	IF_TERMIOS("lnext",	&opt_vlnext)
#if defined(F_SETLKW)
	IF_ANY    ("lock",	&opt_f_setlkw)	/* POSIX, first choice */
#elif defined(HAVE_FLOCK)
	IF_ANY    ("lock",	&opt_flock_ex)	/* BSD, fallback */
#endif
#if defined(F_SETLKW)
	IF_ANY    ("lockw",	&opt_f_setlkw)	/* POSIX, first choice */
#elif defined(HAVE_FLOCK)
	IF_ANY    ("lockw",	&opt_flock_ex_nb)	/* BSD, fallback */
#endif
	IF_EXEC   ("login",	&opt_dash)
	IF_IPAPP  ("lowport",	&opt_lowport)
	IF_ANY    ("lseek",	&opt_seek_set)
#ifdef TCP_MAXSEG
	IF_TCP    ("maxseg",	&opt_tcp_maxseg)
	IF_TCP    ("maxseg-late",	&opt_tcp_maxseg_late)
#endif
	IF_OPENSSL("method",	&opt_openssl_method)
	IF_TERMIOS("min",	&opt_vmin)
	IF_ANY    ("mode",	&opt_perm)
#ifdef TCP_MAXSEG
	IF_TCP    ("mss",	&opt_tcp_maxseg)
	IF_TCP    ("mss-late",	&opt_tcp_maxseg_late)
#endif
#ifdef IP_MTU
	IF_IP     ("mtu",	&opt_ip_mtu)
#endif
#ifdef IP_MTU_DISCOVER
	IF_IP     ("mtudiscover",	&opt_ip_mtu_discover)
#endif
	IF_IP     ("multicastloop",	&opt_ip_multicast_loop)
	IF_IP     ("multicastttl",	&opt_ip_multicast_ttl)
#if defined(O_NDELAY) && (!defined(O_NONBLOCK) || O_NDELAY != O_NONBLOCK)
	IF_ANY    ("ndelay",	&opt_o_ndelay)
#else
	IF_ANY    ("ndelay",	&opt_nonblock)
#endif
	IF_NAMED  ("new",	&opt_unlink_early)
#ifdef NLDLY
#  ifdef NL0
	IF_TERMIOS("nl0",	&opt_nl0)
#  endif
#  ifdef NL1
	IF_TERMIOS("nl1",	&opt_nl1)
#  endif
	IF_TERMIOS("nldly",	&opt_nldly)
#endif /* defined(NLDLY) */
#ifdef SO_NO_CHECK
	IF_SOCKET ("no-check",	&opt_so_no_check)
	IF_SOCKET ("nocheck",	&opt_so_no_check)
#endif
	IF_OPEN   ("noctty",	&opt_o_noctty)
#ifdef TCP_NODELAY
	IF_TCP    ("nodelay",	&opt_tcp_nodelay)
#endif
	IF_TERMIOS("noflsh",	&opt_noflsh)
#ifdef O_NOFOLLOW
	IF_OPEN   ("nofollow",	&opt_o_nofollow)
#endif
	IF_ANY    ("nonblock",	&opt_nonblock)
#ifdef SO_NOREUSEADDR	/* AIX 4.3.3 */
	IF_SOCKET ("noreuseaddr",	&opt_so_noreuseaddr)
#endif /* SO_NOREUSEADDR */
#ifdef O_NSHARE
	IF_OPEN   ("nshare",	&opt_o_nshare)
#endif
	IF_OPEN   ("o-create",	&opt_o_create)
#ifdef O_DEFER
	IF_OPEN   ("o-defer",	&opt_o_defer)
#endif
#ifdef O_DELAY
	IF_OPEN   ("o-delay",	&opt_o_delay)
#endif
#ifdef O_DIRECT
	IF_OPEN   ("o-direct",	&opt_o_direct)
#endif
#ifdef O_DIRECTORY
	IF_OPEN   ("o-directory",	&opt_o_directory)
#endif
#ifdef O_DSYNC
	IF_OPEN   ("o-dsync",	&opt_o_dsync)
#endif
	IF_OPEN   ("o-excl",	&opt_o_excl)
#ifdef O_LARGEFILE
	IF_OPEN   ("o-largefile",	&opt_o_largefile)
#endif
#if defined(O_NDELAY) && (!defined(O_NONBLOCK) || O_NDELAY != O_NONBLOCK)
	IF_ANY    ("o-ndelay",	&opt_o_ndelay)
#else
	IF_ANY    ("o-ndelay",	&opt_nonblock)
#endif
	IF_OPEN   ("o-noctty",	&opt_o_noctty)
#ifdef O_NOFOLLOW
	IF_OPEN   ("o-nofollow",	&opt_o_nofollow)
#endif
#ifdef O_NSHARE
	IF_OPEN   ("o-nshare",	&opt_o_nshare)
#endif
#ifdef O_PRIV
	IF_OPEN   ("o-priv",	&opt_o_priv)
#endif
	IF_OPEN   ("o-rdonly",	&opt_o_rdonly)
	IF_OPEN   ("o-rdwr",	&opt_o_rdwr)
#ifdef O_RSHARE
	IF_OPEN   ("o-rshare",	&opt_o_rshare)
#endif
#ifdef O_RSYNC
	IF_OPEN   ("o-rsync",	&opt_o_rsync)
#endif
#ifdef O_SYNC
	IF_OPEN   ("o-sync",	&opt_o_sync)
#endif
	IF_OPEN   ("o-trunc",	&opt_o_trunc)
	IF_OPEN   ("o-wronly",	&opt_o_wronly)
	IF_OPEN   ("o_create",	&opt_o_create)
#ifdef O_DEFER
	IF_OPEN   ("o_defer",	&opt_o_defer)
#endif
#ifdef O_DELAY
	IF_OPEN   ("o_delay",	&opt_o_delay)
#endif
#ifdef O_DIRECT
	IF_OPEN   ("o_direct",	&opt_o_direct)
#endif
#ifdef O_DIRECTORY
	IF_OPEN   ("o_directory",	&opt_o_directory)
#endif
#ifdef O_DSYNC
	IF_OPEN   ("o_dsync",	&opt_o_dsync)
#endif
	IF_OPEN   ("o_excl",	&opt_o_excl)
#ifdef O_LARGEFILE
	IF_OPEN   ("o_largefile",	&opt_o_largefile)
#endif
#if defined(O_NDELAY) && (!defined(O_NONBLOCK) || O_NDELAY != O_NONBLOCK)
	IF_ANY    ("o_ndelay",	&opt_o_ndelay)
#else
	IF_ANY    ("o_ndelay",	&opt_nonblock)
#endif
	IF_OPEN   ("o_noctty",	&opt_o_noctty)
#ifdef O_NOFOLLOW
	IF_OPEN   ("o_nofollow",	&opt_o_nofollow)
#endif
#ifdef O_NSHARE
	IF_OPEN   ("o_nshare",	&opt_o_nshare)
#endif
#ifdef O_PRIV
	IF_OPEN   ("o_priv",	&opt_o_priv)
#endif
	IF_OPEN   ("o_rdonly",	&opt_o_rdonly)
	IF_OPEN   ("o_rdwr",	&opt_o_rdwr)
#ifdef O_RSHARE
	IF_OPEN   ("o_rshare",	&opt_o_rshare)
#endif
#ifdef O_RSYNC
	IF_OPEN   ("o_rsync",	&opt_o_rsync)
#endif
#ifdef O_SYNC
	IF_OPEN   ("o_sync",	&opt_o_sync)
#endif
	IF_OPEN   ("o_wronly",	&opt_o_wronly)
#ifdef OCRNL
	IF_TERMIOS("ocrnl",	&opt_ocrnl)
#endif
#ifdef OFDEL
	IF_TERMIOS("ofdel",	&opt_ofdel)
#endif
#ifdef OFILL
	IF_TERMIOS("ofill",	&opt_ofill)
#endif
#ifdef OLCUC
	IF_TERMIOS("olcuc",	&opt_olcuc)
#endif
	IF_TERMIOS("onlcr",	&opt_onlcr)
#ifdef ONLRET
	IF_TERMIOS("onlret",	&opt_onlret)
#endif
#ifdef ONOCR
	IF_TERMIOS("onocr",	&opt_onocr)
#endif
	IF_SOCKET ("oobinline",	&opt_so_oobinline)
#if HAVE_OPENPTY
	IF_EXEC   ("openpty",	&opt_openpty)
#endif /* HAVE_OPENPTY */
	IF_OPENSSL("openssl-cipherlist",	&opt_openssl_cipherlist)
	IF_OPENSSL("openssl-method",	&opt_openssl_method)
	IF_OPENSSL("openssl-verify",	&opt_openssl_verify)
	IF_TERMIOS("opost",	&opt_opost)
#ifdef HAVE_TERMIOS_ISPEED
	IF_TERMIOS("ospeed",	&opt_ospeed)
#endif
	IF_ANY    ("owner",	&opt_user)
	IF_TERMIOS("parenb",	&opt_parenb)
	IF_TERMIOS("parmrk",	&opt_parmrk)
	IF_TERMIOS("parodd",	&opt_parodd)
#ifdef SO_PASSCRED
	IF_SOCKET ("passcred",	&opt_so_passcred)
#endif
	IF_EXEC   ("path",	&opt_path)
#ifdef SO_PEERCRED
	IF_SOCKET ("peercred",	&opt_so_peercred)
#endif
#ifdef PENDIN
	IF_TERMIOS("pendin",	&opt_pendin)
#endif
	IF_ANY    ("perm",	&opt_perm)
	IF_NAMED  ("perm-early",	&opt_perm_early)
	IF_ANY    ("perm-late",	&opt_perm_late)
	IF_EXEC   ("pgid",	&opt_setpgid)
	IF_EXEC   ("pipes",	&opt_pipes)
#ifdef IP_PKTINFO
	IF_IP     ("pktinfo",	&opt_ip_pktinfo)
#endif
#ifdef IP_PKTOPTIONS
	IF_IP     ("pktoptions",	&opt_ip_pktoptions)
	IF_IP     ("pktopts",	&opt_ip_pktoptions)
#endif
	/*IF_IPAPP("port",	&opt_port)*/
#ifdef SO_PRIORITY
	IF_SOCKET ("priority",	&opt_so_priority)
#endif
#ifdef O_PRIV
	IF_OPEN   ("priv",	&opt_o_priv)
#endif
#ifdef SO_PROTOTYPE
	IF_SOCKET ("prototype",	&opt_so_prototype)
#endif
	IF_PROXY  ("proxy-authorization",	&opt_proxy_authorization)
	IF_PROXY  ("proxy-auth",	&opt_proxy_authorization)
	IF_PROXY  ("proxy-resolve",	&opt_proxy_resolve)
	IF_PROXY  ("proxyauth",	&opt_proxy_authorization)
	IF_PROXY  ("proxyport",	&opt_proxyport)
#ifdef ECHOPRT
	IF_TERMIOS("prterase",	&opt_echoprt)
#endif
#if HAVE_DEV_PTMX || HAVE_DEV_PTC
	IF_EXEC   ("ptmx",	&opt_ptmx)
#endif
#if HAVE_PTY
	IF_EXEC   ("pty",	&opt_pty)
#endif
#ifdef TCP_QUICKACK
	IF_TCP    ("quickack",	&opt_tcp_quickack)
#endif
	IF_TERMIOS("quit",	&opt_vquit)
	IF_RANGE  ("range",	&opt_range)
	IF_TERMIOS("raw",	&opt_raw)
	IF_SOCKET ("rcvbuf",	&opt_so_rcvbuf)
	IF_SOCKET ("rcvbuf-late",	&opt_so_rcvbuf_late)
#ifdef SO_RCVLOWAT
	IF_SOCKET ("rcvlowat",	&opt_so_rcvlowat)
#endif
#ifdef SO_RCVTIMEO
	IF_SOCKET ("rcvtimeo",	&opt_so_rcvtimeo)
#endif
	IF_OPEN   ("rdonly",	&opt_o_rdonly)
	IF_OPEN   ("rdwr",	&opt_o_rdwr)
#ifdef IP_RECVERR
	IF_IP     ("recverr",	&opt_ip_recverr)
#endif
#ifdef IP_RECVOPTS
	IF_IP     ("recvopts",	&opt_ip_recvopts)
#endif
#ifdef IP_RECVTOS
	IF_IP     ("recvtos",	&opt_ip_recvtos)
#endif
#ifdef IP_RECVTTL
	IF_IP     ("recvttl",	&opt_ip_recvttl)
#endif
	IF_NAMED  ("remove",	&opt_unlink)
	IF_TERMIOS("reprint",	&opt_vreprint)
	IF_PROXY  ("resolv",	&opt_proxy_resolve)
	IF_PROXY  ("resolve",	&opt_proxy_resolve)
#ifdef IP_RETOPTS
	IF_IP     ("retopts",	&opt_ip_retopts)
#endif
	IF_RETRY  ("retry",	&opt_retry)
	IF_SOCKET ("reuseaddr",	&opt_so_reuseaddr)
#ifdef SO_REUSEPORT	/* AIX 4.3.3 */
	IF_SOCKET ("reuseport",	&opt_so_reuseport)
#endif /* defined(SO_REUSEPORT) */
#ifdef TCP_RFC1323
	IF_TCP    ("rfc1323",	&opt_tcp_rfc1323)
#endif
#ifdef IP_ROUTER_ALERT
	IF_IP     ("routeralert",	&opt_ip_router_alert)
#endif
	IF_TERMIOS("rprnt",	&opt_vreprint)
#ifdef O_RSHARE
	IF_OPEN   ("rshare",	&opt_o_rshare)
#endif
#ifdef O_RSYNC
	IF_OPEN   ("rsync",	&opt_o_rsync)
#endif
	IF_TERMIOS("sane",	&opt_sane)
#ifdef SO_SECURITY_AUTHENTICATION
	IF_SOCKET ("security-authentication",	&opt_so_security_authentication)
#endif
#ifdef SO_SECURITY_ENCRYPTION_NETWORK
	IF_SOCKET ("security-encryption-network",	&opt_so_security_encryption_network)
#endif
#ifdef SO_SECURITY_ENCRYPTION_TRANSPORT
	IF_SOCKET ("security-encryption-transport",	&opt_so_security_encryption_transport)
#endif
#ifdef SO_SECURITY_AUTHENTICATION
	IF_SOCKET ("securityauthentication",	&opt_so_security_authentication)
#endif
#ifdef SO_SECURITY_ENCRYPTION_NETWORK
	IF_SOCKET ("securityencryptionnetwork",	&opt_so_security_encryption_network)
#endif
#ifdef SO_SECURITY_ENCRYPTION_TRANSPORT
	IF_SOCKET ("securityencryptiontransport",	&opt_so_security_encryption_transport)
#endif
	IF_ANY    ("seek",	&opt_seek_set)
	IF_ANY    ("seek-cur",	&opt_seek_cur)
	IF_ANY    ("seek-end",	&opt_seek_end)
	IF_ANY    ("seek-set",	&opt_seek_set)
	IF_ANY    ("setgid",	&opt_setgid)
	IF_ANY    ("setgid-early",	&opt_setgid_early)
	IF_ANY 	  ("setlk",	&opt_f_setlk)
	IF_ANY 	  ("setlkw",	&opt_f_setlkw)
	IF_EXEC   ("setpgid",	&opt_setpgid)
#if WITH_EXEC || WITH_SYSTEM
	IF_EXEC   ("setsid",	&opt_setsid)
#endif
	IF_ANY    ("setuid",	&opt_setuid)
	IF_ANY    ("setuid-early",	&opt_setuid_early)
#if WITH_EXEC || WITH_SYSTEM
	IF_ANY    ("sid",	&opt_setsid)
#endif
#ifdef SIOCSPGRP
	IF_SOCKET ("siocspgrp",	&opt_siocspgrp)
#endif
	IF_SOCKET ("sndbuf",	&opt_so_sndbuf)
	IF_SOCKET ("sndbuf-late",	&opt_so_sndbuf_late)
#ifdef SO_SNDLOWAT
	IF_SOCKET ("sndlowat",	&opt_so_sndlowat)
#endif
#ifdef SO_SNDTIMEO
	IF_SOCKET ("sndtimeo",	&opt_so_sndtimeo)
#endif
#ifdef SO_ACCEPTCONN /* AIX433 */
	IF_SOCKET ("so-acceptconn",	&opt_so_acceptconn)
#endif /* SO_ACCEPTCONN */
#ifdef SO_ATTACH_FILTER
	IF_SOCKET ("so-attach-filter",	&opt_so_attach_filter)
#endif
#ifdef SO_AUDIT	/* AIX 4.3.3 */
	IF_SOCKET ("so-audit",	&opt_so_audit)
#endif /* SO_AUDIT */
#ifdef SO_BINDTODEVICE
	IF_SOCKET ("so-bindtodevice",	&opt_so_bindtodevice)
#endif
	IF_SOCKET ("so-broadcast",	&opt_so_broadcast)
#ifdef SO_BSDCOMPAT
	IF_SOCKET ("so-bsdcompat",	&opt_so_bsdcompat)
#endif
#ifdef SO_CKSUMRECV
	IF_SOCKET ("so-cksumrecv",	&opt_so_cksumrecv)
#endif /* SO_CKSUMRECV */
	IF_SOCKET ("so-debug",	&opt_so_debug)
#ifdef SO_DETACH_FILTER
	IF_SOCKET ("so-detach-filter", &opt_so_detach_filter)
#endif
#ifdef SO_DGRAM_ERRIND
	IF_SOCKET ("so-dgram-errind",	&opt_so_dgram_errind)
#endif
#ifdef SO_DONTLINGER
	IF_SOCKET ("so-dontlinger",	&opt_so_dontlinger)
#endif
	IF_SOCKET ("so-dontroute",	&opt_so_dontroute)
	IF_SOCKET ("so-error",	&opt_so_error)
	IF_SOCKET ("so-linger",	&opt_so_linger)
	IF_SOCKET ("so-keepalive",	&opt_so_keepalive)
#ifdef SO_KERNACCEPT	/* AIX 4.3.3 */
	IF_SOCKET ("so-kernaccept",	&opt_so_kernaccept)
#endif /* SO_KERNACCEPT */
#ifdef SO_NO_CHECK
	IF_SOCKET ("so-no-check",	&opt_so_no_check)
#endif
#ifdef SO_NOREUSEADDR	/* AIX 4.3.3 */
	IF_SOCKET ("so-noreuseaddr",	&opt_so_noreuseaddr)
#endif /* SO_NOREUSEADDR */
	IF_SOCKET ("so-oobinline",	&opt_so_oobinline)
#ifdef SO_PASSCRED
	IF_SOCKET ("so-passcred",	&opt_so_passcred)
#endif
#ifdef SO_PEERCRED
	IF_SOCKET ("so-peercred",	&opt_so_peercred)
#endif
#ifdef SO_PRIORITY
	IF_SOCKET ("so-priority",	&opt_so_priority)
#endif
#ifdef SO_PROTOTYPE
	IF_SOCKET ("so-prototype",	&opt_so_prototype)
#endif
	IF_SOCKET ("so-rcvbuf",	&opt_so_rcvbuf)
	IF_SOCKET ("so-rcvbuf-late",	&opt_so_rcvbuf_late)
#ifdef SO_RCVLOWAT
	IF_SOCKET ("so-rcvlowat",	&opt_so_rcvlowat)
#endif
#ifdef SO_RCVTIMEO
	IF_SOCKET ("so-rcvtimeo",	&opt_so_rcvtimeo)
#endif
	IF_SOCKET ("so-reuseaddr",	&opt_so_reuseaddr)
#ifdef SO_REUSEPORT	/* AIX 4.3.3 */
	IF_SOCKET ("so-reuseport",	&opt_so_reuseport)
#endif /* defined(SO_REUSEPORT) */
#ifdef SO_SECURITY_AUTHENTICATION
	IF_SOCKET ("so-security-authentication",	&opt_so_security_authentication)
#endif
#ifdef SO_SECURITY_ENCRYPTION_NETWORK
	IF_SOCKET ("so-security-encryption-network",	&opt_so_security_encryption_network)
#endif
#ifdef SO_SECURITY_ENCRYPTION_TRANSPORT
	IF_SOCKET ("so-security-encryption-transport",	&opt_so_security_encryption_transport)
#endif
	IF_SOCKET ("so-sndbuf",	&opt_so_sndbuf)
	IF_SOCKET ("so-sndbuf-late",	&opt_so_sndbuf_late)
#ifdef SO_SNDLOWAT
	IF_SOCKET ("so-sndlowat",	&opt_so_sndlowat)
#endif
#ifdef SO_SNDTIMEO
	IF_SOCKET ("so-sndtimeo",	&opt_so_sndtimeo)
#endif
	IF_SOCKET ("so-type",	&opt_so_type)
#ifdef SO_USE_IFBUFS
	IF_SOCKET ("so-use-ifbufs",	&opt_so_use_ifbufs)
#endif /* SO_USE_IFBUFS */
#ifdef SO_USELOOPBACK /* AIX433, Solaris */
	IF_SOCKET ("so-useloopback",	&opt_so_useloopback)
#endif /* SO_USELOOPBACK */
 	IF_SOCKS4 ("socksport",	&opt_socksport)
	IF_SOCKS4 ("socksuser",	&opt_socksuser)
	IF_IPAPP  ("sourceport",	&opt_sourceport)
	IF_IPAPP  ("sp",	&opt_sourceport)
	IF_TERMIOS("start",	&opt_vstart)
	IF_EXEC   ("stderr",    &opt_stderr)
#ifdef TCP_STDURG
	IF_TCP    ("stdurg",	&opt_tcp_stdurg)
#endif
	IF_TERMIOS("stop",	&opt_vstop)
	IF_ANY    ("su",	&opt_substuser)
	IF_ANY    ("su-d",	&opt_substuser_delayed)
	IF_ANY    ("substuser",	&opt_substuser)
	IF_ANY    ("substuser-delayed",	&opt_substuser_delayed)
	IF_TERMIOS("susp",	&opt_vsusp)
#ifdef VSWTC
	IF_TERMIOS("swtc",	&opt_vswtc)
	IF_TERMIOS("swtch",	&opt_vswtc)
#endif
	IF_PTY    ("symbolic-link",	&opt_symbolic_link)
#ifdef O_SYNC
	IF_OPEN   ("sync",	&opt_o_sync)
#endif
#ifdef TCP_SYNCNT
	IF_TCP    ("syncnt",	&opt_tcp_syncnt)
#endif
#ifdef TABDLY
#  ifdef TAB0
	IF_TERMIOS("tab0",	&opt_tab0)
#  endif
#  ifdef TAB1
	IF_TERMIOS("tab1",	&opt_tab1)
#  endif
#  ifdef TAB2
	IF_TERMIOS("tab2",	&opt_tab2)
#  endif
#  ifdef TAB3
	IF_TERMIOS("tab3",	&opt_tab3)
#  endif
	IF_TERMIOS("tabdly",	&opt_tabdly)
#endif
	IF_TERMIOS("tandem",	&opt_ixoff)
#ifdef TCP_CORK
	IF_TCP    ("tcp-cork",	&opt_tcp_cork)
#endif
#ifdef TCP_DEFER_ACCEPT	/* Linux 2.4.0 */
	IF_TCP    ("tcp-defer-accept",	&opt_tcp_defer_accept)
#endif
#ifdef TCP_INFO	/* Linux 2.4.0 */
	IF_TCP    ("tcp-info",	&opt_tcp_info)
#endif
#ifdef TCP_KEEPCNT	/* Linux 2.4.0 */
	IF_TCP    ("tcp-keepcnt",	&opt_tcp_keepcnt)
#endif
#ifdef TCP_KEEPIDLE	/* Linux 2.4.0 */
	IF_TCP    ("tcp-keepidle",	&opt_tcp_keepidle)
#endif
#ifdef TCP_KEEPINTVL	/* Linux 2.4.0 */
	IF_TCP    ("tcp-keepintvl",	&opt_tcp_keepintvl)
#endif
#ifdef TCP_LINGER2	/* Linux 2.4.0 */
	IF_TCP    ("tcp-linger2",	&opt_tcp_linger2)
#endif
#ifdef TCP_MAXSEG
	IF_TCP    ("tcp-maxseg",	&opt_tcp_maxseg)
	IF_TCP    ("tcp-maxseg-late",	&opt_tcp_maxseg_late)
#endif
#ifdef TCP_NODELAY
	IF_TCP    ("tcp-nodelay",	&opt_tcp_nodelay)
#endif
#ifdef TCP_QUICKACK
	IF_TCP    ("tcp-quickack",	&opt_tcp_quickack)
#endif
#ifdef TCP_RFC1323
	IF_TCP    ("tcp-rfc1323",	&opt_tcp_rfc1323)
#endif
#ifdef TCP_STDURG
	IF_TCP    ("tcp-stdurg",	&opt_tcp_stdurg)
#endif
#ifdef TCP_SYNCNT	/* Linux 2.4.0 */
	IF_TCP    ("tcp-syncnt",	&opt_tcp_syncnt)
#endif
#ifdef TCP_WINDOW_CLAMP	/* Linux 2.4.0 */
	IF_TCP    ("tcp-window-clamp",	&opt_tcp_window_clamp)
#endif
#if WITH_LIBWRAP
	IF_IPAPP  ("tcpwrap",		&opt_tcpwrappers)
	IF_IPAPP  ("tcpwrapper",	&opt_tcpwrappers)
	IF_IPAPP  ("tcpwrappers",	&opt_tcpwrappers)
#endif
	IF_TERMIOS("time",	&opt_vtime)
	IF_TERMIOS("tiocsctty",	&opt_tiocsctty)
	IF_IP     ("tos",	&opt_ip_tos)
	IF_TERMIOS("tostop",	&opt_tostop)
	IF_OPEN   ("trunc",	&opt_o_trunc)
	IF_ANY    ("truncate",	&opt_ftruncate)
	IF_IP     ("ttl",	&opt_ip_ttl)
	IF_SOCKET ("type",	&opt_so_type)
	IF_ANY    ("uid",	&opt_user)
	IF_NAMED  ("uid-e",	&opt_user_early)
	IF_ANY    ("uid-l",	&opt_user_late)
	IF_NAMED  ("umask",	&opt_umask)
	IF_NAMED  ("unlink",	&opt_unlink)
	IF_NAMED  ("unlink-early",	&opt_unlink_early)
	IF_NAMED  ("unlink-late",	&opt_unlink_late)
#ifdef SO_USE_IFBUFS
	IF_SOCKET ("use-ifbufs",	&opt_so_use_ifbufs)
	IF_SOCKET ("useifbufs",		&opt_so_use_ifbufs)
#endif /* SO_USE_IFBUFS */
#ifdef SO_USELOOPBACK /* AIX433, Solaris */
	IF_SOCKET ("useloopback",	&opt_so_useloopback)
#endif /* SO_USELOOPBACK */
	IF_ANY    ("user",	&opt_user)
	IF_NAMED  ("user-early",	&opt_user_early)
	IF_ANY    ("user-late",	&opt_user_late)
#ifdef VDISCARD
	IF_TERMIOS("vdiscard",	&opt_vdiscard)
#endif
	IF_TERMIOS("veof",	&opt_veof)
	IF_TERMIOS("veol",	&opt_veol)
	IF_TERMIOS("veol2",	&opt_veol2)
	IF_TERMIOS("verase",	&opt_verase)
	IF_OPENSSL("verify",	&opt_openssl_verify)
	IF_TERMIOS("vintr",	&opt_vintr)
	IF_TERMIOS("vkill",	&opt_vkill)
	IF_TERMIOS("vlnext",	&opt_vlnext)
	IF_TERMIOS("vmin",	&opt_vmin)
	IF_TERMIOS("vquit",	&opt_vquit)
	IF_TERMIOS("vreprint",	&opt_vreprint)
	IF_TERMIOS("vstart",	&opt_vstart)
	IF_TERMIOS("vstop",	&opt_vstop)
	IF_TERMIOS("vsusp",	&opt_vsusp)
#ifdef VSWTC
	IF_TERMIOS("vswtc",	&opt_vswtc)
#endif
#ifdef VTDLY
#  ifdef VT0
	IF_TERMIOS("vt0",	&opt_vt0)
#  endif
#  ifdef VT1
	IF_TERMIOS("vt1",	&opt_vt1)
#  endif
	IF_TERMIOS("vtdly",	&opt_vtdly)
#endif
	IF_TERMIOS("vtime",	&opt_vtime)
#ifdef VWERASE
	IF_TERMIOS("vwerase",	&opt_vwerase)
#endif
#ifdef VWERASE
	IF_TERMIOS("werase",	&opt_vwerase)
#endif
#ifdef TCP_WINDOW_CLAMP	/* Linux 2.4.0 */
	IF_TCP    ("window-clamp",	&opt_tcp_window_clamp)
#endif
#if WITH_LIBWRAP
	IF_IPAPP  ("wrap",		&opt_tcpwrappers)
#endif
	IF_OPEN   ("wronly",	&opt_o_wronly)
#ifdef XCASE
	IF_TERMIOS("xcase",	&opt_xcase)
#endif
#if defined(TABDLY) && defined(XTABS)
	IF_TERMIOS("xtabs",	&opt_xtabs)
#endif
	{ NULL }
} ;


/* walks the text argument a and writes its options that conform to groups 
   to the array opts
   returns 0 on success, -1 on error, 1 on unknown/wrong option  
*/
int parseopts(char *a, unsigned int groups, struct opt **opts) {
   int i=0;
   char *equal, *comma = a, *a0, *a2, *pisep;
   unsigned long ulongval;
   long slongval;
   char optr = ',';
   int result;
   byte_t optbuf[256];  size_t optlen;

   *opts = Malloc((i+8)*sizeof(struct opt));
   if (*opts == NULL) {
      return -1;
   }
   if (a == NULL) {
      (*opts)[i].desc = ODESC_END;
      return 0;
   }

   while (optr == ',' || optr == '=') {
      const struct optname *ent;

      if (a == NULL || *a == '\0')
	 break;

      if (*a == ',') ++a;
      a0 = a;
      equal = strchr(a, '=');
      comma = strchr(a, ',');
      pisep = strchr(a, xioopts.pipesep[0]);
      if (comma == NULL) {
	 comma = pisep;
      } else if (pisep != NULL) {
	 comma = Min(comma, pisep);
      }
      /* find the termination of the next option name:
	 is nearest of '=' or ',' or '\0' */
      if (equal != NULL) {
	 if (comma != NULL) {
	    if (equal < comma) {
	       optr = *equal;  *equal = *comma++ = '\0'; a2 = equal+1;
	    } else {
	       optr = *comma;  *comma++ = '\0'; a2 = NULL;
	    }
	 } else {
	    optr = *equal;  *equal = '\0'; a2 = equal+1;
	 }
      } else {
	 if (comma != NULL) {
	    optr = *comma;  *comma++ = '\0'; a2 = NULL;
	 } else {
	    optr = ','; a2 = NULL;
	 }
      }
      a = comma;

      ent = (struct optname *)
	 keyw((struct wordent *)optionnames, a0,
	      sizeof(optionnames)/sizeof(struct optname)-1);
      if (ent == NULL) {
	 Error1("parseopts(): unknown option \"%s\"", a0);
	 continue;
      }
      if (!(ent->desc->group & groups) && !(ent->desc->group & GROUP_ANY) &&
	  !xioopts_ignoregroups) {
	 Error1("parseopts(): option \"%s\" not supported with this address type",
		a0);
	 Error2("parseopts()  groups=%08x, ent->group=%08x",
	       groups, ent->desc->group);
	 continue;
      }
      (*opts)[i].desc = ent->desc;
      switch (ent->desc->type) {
      case TYPE_CONST:
	 if (optr == '=') {
	    Error1("no value permitted for option \"%s\"",
		   ent->desc->defname);
	    continue;
	 }
	 Info1("setting option \"%s\"", ent->desc->defname);
	 break;
      case TYPE_BIN:
	 if (a2 == NULL) { Error1("option \"%s\": value required", a0);
	    continue; }
	 optlen = 0;
	 if ((result = dalan(a2, optbuf, &optlen, sizeof(optbuf))) != 0) {
	    Error1("parseopts(): problem with \"%s\" data", a2);
	    continue;
	 }
	 if (((*opts)[i].value.u_bin.b_data = memdup(optbuf, optlen)) == NULL) {
	    Error1("memdup(, "F_Zu"): out of memory", optlen);
	    return -1;
	 }
	 (*opts)[i].value.u_bin.b_len = optlen;
	 break;
      case TYPE_BYTE:
	 if (a2) {
	    (*opts)[i].value.u_byte = strtoul(a2, &a2, 0);
	 } else {
	    (*opts)[i].value.u_byte = 1;
	 }
	 if ((*opts)[i].value.u_byte > UCHAR_MAX) {
	    Error3("parseopts(%s): byte value exceeds limit (%u vs. %u), using max",
		   a0, (*opts)[i].value.u_byte, UCHAR_MAX);
	 } else {
	    Info2("setting option \"%s\" to %d", ent->desc->defname,
		  (*opts)[i].value.u_byte);
	 }
	 break;
      case TYPE_INT:
	 if (a2) {
	    (*opts)[i].value.u_int = strtoul(a2, &a2, 0);
	 } else {
	    (*opts)[i].value.u_int = 1;
	 }
	 Info2("setting option \"%s\" to %d", ent->desc->defname,
	       (*opts)[i].value.u_int);
	 break;
      case TYPE_BOOL:
	 if (optr != '=') {
	    (*opts)[i].value.u_bool = 1;
	 } else {
	    (*opts)[i].value.u_bool = strtoul(a2, &a2, 0);
	 }
	 Info2("setting option \"%s\" to %d", ent->desc->defname,
	       (*opts)[i].value.u_bool);
	 break;
      case TYPE_UINT:
	 if (optr != '=') {
	    (*opts)[i].value.u_uint = 1;
	 } else {
	    ulongval = strtoul(a2, &a2, 0);
	    if (ulongval > UINT_MAX) {
	       Error3("parseopts(%s): unsigned int value exceeds limit (%u vs. %u), using max",
		      a0, ulongval, UINT_MAX);
	    }
	    (*opts)[i].value.u_uint = ulongval;
	 }
	 Info2("setting option \"%s\" to %u", ent->desc->defname,
	       (*opts)[i].value.u_uint);
	 break;
      case TYPE_USHORT:
	 if (optr != '=') {
	    (*opts)[i].value.u_ushort = 1;
	 } else {
	    ulongval = strtoul(a2, &a2, 0);
	    if (ulongval > USHRT_MAX) {
	       Error3("parseopts(%s): unsigned short value exceeds limit (%u vs. %u), using max",
		      a0, ulongval, USHRT_MAX);
	    }
	    (*opts)[i].value.u_ushort = ulongval;
	 }
	 Info2("setting option \"%s\" to %u", ent->desc->defname,
	       (*opts)[i].value.u_ushort);
	 break;
      case TYPE_LONG:
	 if (optr != '=') {
	    (*opts)[i].value.u_long = 1;
	 } else {
	    slongval = strtol(a2, &a2, 0);
	    (*opts)[i].value.u_long = slongval;
	 }
	 Info2("setting option \"%s\" to %u", ent->desc->defname,
	       (*opts)[i].value.u_long);
	 break;
      case TYPE_UIDT:
	 if (a2 == NULL) { Error1("option \"%s\": value required", a0);
	    continue; }
	 if (isdigit((*a2)&0xff)) {
	    (*opts)[i].value.u_uidt = strtoul(a2, &a2, 0);
	 } else {
	    struct passwd *pwd;
	    if ((pwd = getpwnam(a2)) == NULL) {
	       Error1("getpwnam(\"%s\"): no such user", a2);
	       continue;
	    }
	    (*opts)[i].value.u_uidt = getpwnam(a2)->pw_uid;
	 }
	 Info2("setting option \"%s\" to %u", ent->desc->defname,
	       (*opts)[i].value.u_uidt);
	 break;
      case TYPE_GIDT:
	 if (a2 == NULL) { Error1("option \"%s\": value required", a0);
	    continue; }
	 if (isdigit((*a2)&0xff)) {
	    (*opts)[i].value.u_gidt = strtoul(a2, &a2, 0);
	 } else {
	    struct group *grp;
	    grp = getgrnam(a2);
	    if (grp == NULL) {
	       Error1("getgrnam(\"%s\"): no such group", a2);
	       continue;
	    }
	    (*opts)[i].value.u_gidt = grp->gr_gid;
	 }
	 Info2("setting option \"%s\" to %u", ent->desc->defname,
	       (*opts)[i].value.u_gidt);
	 break;
      case TYPE_MODET:
	 if (optr != '=') { Error1("option \"%s\": value required", a0);
	    continue;
	 }
	 (*opts)[i].value.u_modet = strtoul(a2, &a2, 8);
	 Info2("setting option \"%s\" to %u", ent->desc->defname,
	       (*opts)[i].value.u_modet);
	 break;
      case TYPE_STRING:
	 if (optr != '=') {
	    Error1("option \"%s\": value required", a0);
	    continue;
	 }
	 (*opts)[i].value.u_string = a2;
	 Info2("setting option \"%s\" to \"%s\"", ent->desc->defname,
	       (*opts)[i].value.u_string);
	 break;
      case TYPE_STRING_NULL:
	 if (optr != '=') {
	    (*opts)[i].value.u_string = NULL;
	    Info1("setting option \"%s\" to NULL", ent->desc->defname);
	 } else {
	    (*opts)[i].value.u_string = a2;
	    Info2("setting option \"%s\" to \"%s\"", ent->desc->defname,
		  (*opts)[i].value.u_string);
	 }
	 break;
#if LATER
      case TYPE_INT3:
	 
	 break;
#endif
      case TYPE_TIMEVAL:
	 if (optr != '=') {
	    Error1("option \"%s\": value required", a0);
	    continue;
	 } else {
	    double val;
	    val = strtod(a2, NULL);
	    if (val == HUGE_VAL || val == -HUGE_VAL ||
		val == 0.0 && errno == ERANGE) {
	       Error2("strtod(\"%s\", NULL): %s", a2, strerror(errno));
	       val = 0.0;
	    }
	    (*opts)[i].value.u_timeval.tv_sec  = val;
	    (*opts)[i].value.u_timeval.tv_usec =
	       (val-(*opts)[i].value.u_timeval.tv_sec) * 1000000;
	 }
	 break;
#if STRUCT_LINGER
      case TYPE_LINGER:
	 if (optr != '=') {
	    Error1("option \"%s\": value required", a0);
	    continue;
	 }
	 (*opts)[i].value.u_linger.l_onoff = 1;
	 (*opts)[i].value.u_linger.l_linger = strtoul(a2, &a2, 0);
	 Info3("setting option \"%s\" to {%d,%d}", ent->desc->defname,
	       (*opts)[i].value.u_linger.l_onoff,
	       (*opts)[i].value.u_linger.l_linger);
	 break;
#endif /* STRUCT_LINGER */
      default:
	 Error2("parseopts(): internal error on option \"%s\": unimplemented type %d",
		ent->desc->defname, ent->desc->type);
	 continue;
      }
      if (comma) {
	 a = comma;
      } else {
	 optr = '\0';
      }

      ++i;
      if ((i % 8) == 0) {
	 *opts = Realloc(*opts, (i+8) * sizeof(struct opt));
	 if (*opts == NULL) {
	    return -1;
	 }
      }
   }
   /*(*opts)[i+1].desc = ODESC_END;*/
   (*opts)[i].desc = ODESC_END;
   return 0;
}

/* copy the already parsed options for repeated application, but only those
   matching groups ANY and <groups> */
struct opt *copyopts(const struct opt *opts, unsigned int groups) {
   struct opt *new;
   int i, j, n;

   if (!opts)  return NULL;

   /* just count the options in the array */
   i = 0; while (opts[i].desc != ODESC_END) {
      ++i;
   }
   n = i+1;

   new = Malloc(n * sizeof(struct opt));
   if (new == NULL) {
      return NULL;
   }

   /*0 memcpy(new, opts, (i+1)*sizeof(struct opt)); */
   i = 0, j = 0;
   while (i < n-1) {
      if (opts[i].desc == ODESC_DONE) {
	 new[j].desc = ODESC_DONE;
      } else if ((opts[i].desc->group & (GROUP_ANY&~GROUP_PROCESS)) ||
		 (opts[i].desc->group & groups)) {
	 new[j++] = opts[i];
      }
      ++i;
   }
   new[j].desc = ODESC_END;
   return new;
}

/* move options to a new options list
   move only those matching <groups> */
struct opt *moveopts(struct opt *opts, unsigned int groups) {
   struct opt *new;
   int i, j, n;

   if (!opts)  return NULL;

   /* just count the options in the array */
   i = 0; j = 0; while (opts[i].desc != ODESC_END) {
      if (opts[i].desc != ODESC_DONE &&
	  opts[i].desc != ODESC_ERROR)
	 ++j;
      ++i;
   }
   n = i;

   new = Malloc((j+1) * sizeof(struct opt));
   if (new == NULL) {
      return NULL;
   }

   /*0 memcpy(new, opts, (i+1)*sizeof(struct opt)); */
   i = 0, j = 0;
   while (i < n) {
      if (opts[i].desc == ODESC_DONE ||
	  opts[i].desc == ODESC_ERROR) {
	 ++i; continue;
      } else if (opts[i].desc->group & groups) {
	 new[j++] = opts[i];
	 opts[i].desc = ODESC_DONE;
      }
      ++i;
   }
   new[j].desc = ODESC_END;
   return new;
}

/* return the number of yet unconsumed options; -1 on error */
int leftopts(const struct opt *opts) {
   const struct opt *opt = opts;
   int num = 0;

   if (!opts)  return 0;

   while (opt->desc != ODESC_END) {
      if (opt->desc != ODESC_DONE) {
	 ++num;
      }
      ++opt;
   }
   return num;
}

/* show as warning which options are still unused */
int showleft(const struct opt *opts) {
   const struct opt *opt = opts;

   while (opt->desc != ODESC_END) {
      if (opt->desc != ODESC_DONE) {
	 Warn1("showleft(): option \"%s\" not inquired", opt->desc->defname);
      }
      ++opt;
   }
   return 0;
}


/* determines the address group from mode_t */
/* does not set GROUP_FD; cannot determine GROUP_TERMIOS ! */
int _groupbits(mode_t mode) {
   unsigned int result = 0;
  
   switch ((mode&S_IFMT)>>12) {
   case (S_IFIFO>>12):	/* 1, FIFO */
      result = GROUP_FIFO;   break;
   case (S_IFCHR>>12):	/* 2, character device */
      result = GROUP_CHR|GROUP_TERMIOS;    break;
   case (S_IFDIR>>12):	/* 4, directory !!! not supported */
      result = GROUP_NONE;   break;
   case (S_IFBLK>>12):	/* 6, block device */
      result = GROUP_BLK;    break;
   case (S_IFREG>>12):	/* 8, regular file */
      result = GROUP_REG;    break;
   case (S_IFLNK>>12):	/* 10, symbolic link !!! not supported */
      result = GROUP_NONE;   break;
   case (S_IFSOCK>>12): /* 12, socket */
      result = GROUP_SOCKET|GROUP_SOCK_UNIX; break;
   }
   Debug2("_groupbits(%d) -> %d", mode, result);
   return result;
}

/* does not set GROUP_FD */
int groupbits(int fd) {
   struct stat buf;
   int result;

   if (Fstat(fd, &buf) < 0) {
      Error4("groupbits(%d): fstat(%d, %p): %s",
	     fd, fd, &buf, strerror(errno));
      return -1;
   }
   result = _groupbits(buf.st_mode&S_IFMT);
   if (result == GROUP_CHR) {
      if (Isatty(fd) > 0) {
	 result |= GROUP_TERMIOS;
      }
   }
   return result;
}

#if 0	/* currently not used */
int retropt(struct opt *opts, int optcode, union integral *result) {
   struct opt *opt = opts;

   while (opt->desc != ODESC_END) {
      if (opt->desc != ODESC_DONE && opt->desc->optcode == optcode) {
	 *result = opt->value;
	 opt->desc = ODESC_DONE;
	 return 0;
      }
      ++opt;
   }
   return -1;
}
#endif

/* Looks for the first option of type optcode. If the option is found,
   this function stores its bool value in *result, "consumes" the
   option, and returns 0.
   If the option is not found, *result is not modified, and -1 is returned. */
int retropt_bool(struct opt *opts, int optcode, bool *result) {
   struct opt *opt = opts;

   while (opt->desc != ODESC_END) {
      if (opt->desc != ODESC_DONE && opt->desc->optcode == optcode) {
	 *result = opt->value.u_bool;
	 opt->desc = ODESC_DONE;
	 return 0;
      }
      ++opt;
   }
   return -1;
}

#if 0	/* currently not used */
/* Looks for the first option of type optcode. If the option is found,
   this function stores its short value in *result, "consumes" the
   option, and returns 0.
   If the option is not found, *result is not modified, and -1 is returned. */
int retropt_short(struct opt *opts, int optcode, short *result) {
   struct opt *opt = opts;

   while (opt->desc != ODESC_END) {
      if (opt->desc != ODESC_DONE && opt->desc->optcode == optcode) {
	 *result = opt->value.u_short;
	 opt->desc = ODESC_DONE;
	 return 0;
      }
      ++opt;
   }
   return -1;
}
#endif

/* Looks for the first option of type optcode. If the option is found,
   this function stores its unsigned short value in *result, "consumes" the
   option, and returns 0.
   If the option is not found, *result is not modified, and -1 is returned. */
int retropt_ushort(struct opt *opts, int optcode, unsigned short *result) {
   struct opt *opt = opts;

   while (opt->desc != ODESC_END) {
      if (opt->desc != ODESC_DONE && opt->desc->optcode == optcode) {
	 *result = opt->value.u_ushort;
	 opt->desc = ODESC_DONE;
	 return 0;
      }
      ++opt;
   }
   return -1;
}

/* Looks for the first option of type optcode. If the option is found,
   this function stores its int value in *result, "consumes" the
   option, and returns 0.
   If the option is not found, *result is not modified, and -1 is returned. */
int retropt_int(struct opt *opts, int optcode, int *result) {
   struct opt *opt = opts;

   while (opt->desc != ODESC_END) {
      if (opt->desc != ODESC_DONE && opt->desc->optcode == optcode) {
	 *result = opt->value.u_int;
	 opt->desc = ODESC_DONE;
	 return 0;
      }
      ++opt;
   }
   return -1;
}

/* Looks for the first option of type optcode. If the option is found,
   this function stores its unsigned int value in *result, "consumes" the
   option, and returns 0.
   If the option is not found, *result is not modified, and -1 is returned. */
int retropt_uint(struct opt *opts, int optcode, unsigned int *result) {
   struct opt *opt = opts;

   while (opt->desc != ODESC_END) {
      if (opt->desc != ODESC_DONE && opt->desc->optcode == optcode) {
	 *result = opt->value.u_uint;
	 opt->desc = ODESC_DONE;
	 return 0;
      }
      ++opt;
   }
   return -1;
}

/* Looks for the first option of type optcode. If the option is found,
   this function stores its long value in *result, "consumes" the option,
   and returns 0.
   If the option is not found, *result is not modified, and -1 is returned. */
int retropt_long(struct opt *opts, int optcode, long *result) {
   struct opt *opt = opts;

   while (opt->desc != ODESC_END) {
      if (opt->desc != ODESC_DONE && opt->desc->optcode == optcode) {
	 *result = opt->value.u_long;
	 opt->desc = ODESC_DONE;
	 return 0;
      }
      ++opt;
   }
   return -1;
}

/* Looks for the first option of type optcode. If the option is found,
   this function stores its unsigned long value in *result, "consumes" the
   option, and returns 0.
   If the option is not found, *result is not modified, and -1 is returned. */
int retropt_ulong(struct opt *opts, int optcode, unsigned long *result) {
   struct opt *opt = opts;

   while (opt->desc != ODESC_END) {
      if (opt->desc != ODESC_DONE && opt->desc->optcode == optcode) {
	 *result = opt->value.u_ulong;
	 opt->desc = ODESC_DONE;
	 return 0;
      }
      ++opt;
   }
   return -1;
}

#if 0	/* currently not used */
/* get the value of a FLAG typed option, and apply it to the appropriate
   bit position. Mark the option as consumed (done). return 0 if options was found and successfully applied,
   or -1 if option was not in opts */
int retropt_flag(struct opt *opts, int optcode, flags_t *result) {
   struct opt *opt = opts;

   while (opt->desc != ODESC_END) {
      if (opt->desc != ODESC_DONE && opt->desc->optcode == optcode) {
	 if (opt->value.u_bool) {
	    *result |= opt->desc->major;
	 } else {
	    *result &= ~opt->desc->major;
	 }
	 opt->desc = ODESC_DONE;
	 return 0;
      }
      ++opt;
   }
   return -1;
}
#endif

/* Looks for the first option of type optcode. If the option is found,
   this function stores its character pointer value in *result, "consumes" the
   option, and returns 0. Note that, for options of type STRING_NULL, the
   character pointer might degenerate to NULL.
   If the option is not found, *result is not modified, and -1 is returned. */
int retropt_string(struct opt *opts, int optcode, char **result) {
   struct opt *opt = opts;

   while (opt->desc != ODESC_END) {
      if (opt->desc != ODESC_DONE && opt->desc->optcode == optcode) {
	 if (opt->value.u_string == NULL) {
	    *result = NULL;
	 } else if ((*result = strdup(opt->value.u_string)) == NULL) {
	    Error1("strdup("F_Zu"): out of memory",
		   strlen(opt->value.u_string));
	    return -1;
	 }
	 opt->desc = ODESC_DONE;
	 return 0;
      }
      ++opt;
   }
   return -1;
}


#if WITH_SOCKET
int retropt_bind(struct opt *opts,
		 int af,
		 const char *protoname,	/* for service resolution */
		 struct sockaddr *sa,
		 int feats	/* TCP etc: 1..address allowed,
				   3..address and port allowed */
		 ) {
   char *bindname;
   /*0 struct hostent *usaddr;*/

   if (retropt_string(opts, OPT_BIND, &bindname) < 0) {
      return -1;
   }

   switch (af) {
#if WITH_IP4
   case AF_INET:
      return
	 parseip4addr(bindname, (struct sockaddr_in *)sa, protoname,
		      0, (feats>=2)?0:-1);
      break;
#endif /* WITH_IP4 */

#if WITH_UNIX
   case AF_UNIX:
      {
	 struct sockaddr_un *s_un = (struct sockaddr_un *)sa;
	 strncpy(s_un->sun_path, bindname, UNIX_PATH_MAX);
      }
      break;
#endif /* WITH_UNIX */

#if WITH_IP6
   case AF_INET6:
      switch (bindname[0]) {
	 char *np;
	 struct sockaddr_in6 *sin6;
      case '.': case ':': /* only port */
	 sin6 = (struct sockaddr_in6 *) sa;
	 if (!(feats & 2)) {
	    Error("bind cannot set port here");
	    return -1;
	 } else {
	    sin6 = (struct sockaddr_in6 *)sa;
	    sin6->sin6_port = htons(strtoul(bindname+1, &np, 0));
	    if (*np) {
	      Error1("option \"bind\": port number with extra data \"%s\"",
		     np);
	    }
	 }
	 break;
      default:
	 sin6 = (struct sockaddr_in6 *) sa;
	 return parseip6addr(bindname, sin6, protoname,
			     (feats&1)?9:-1, (feats&2)?0:-1);
      }
      break;
#endif /* WITH_IP6 */
   default:
      Error1("bind: unknown address family type %d", af);
      return -1;
   }
   return 0;
}
#endif /* WITH_SOCKET */


/* applies to fd all options belonging to phase */
/* note: not all options can be applied this way (e.g. OFUNC_SPEC with PH_OPEN)
   implemented are: OFUNC_FCNTL, OFUNC_SOCKOPT (probably not all types),
   OFUNC_TERMIOS_FLAG, OFUNC_TERMIOS_PATTERN, and some OFUNC_SPEC */
int applyopts(int fd, struct opt *opts, unsigned int phase) {
   struct opt *opt;

   opt = opts; while (opt && opt->desc != ODESC_END) {
      if (opt->desc == ODESC_DONE ||
	  (phase != PH_ALL && opt->desc->phase != phase)) {
	 ++opt; continue; }

      if (opt->desc->func == OFUNC_SEEK) {
	 if (Lseek(fd, opt->value.u_long, opt->desc->major) < 0) {
	    Error4("lseek(%d, %ld, %d): %s",
		   fd, opt->value.u_long, opt->desc->major, strerror(errno));
	 }

      } else if (opt->desc->func == OFUNC_FCNTL) {
	 int flag;

	 /* retrieve existing flag setttings */
	 if ((flag = Fcntl(fd, opt->desc->major-1)) < 0) {
	    Error3("fcntl(%d, %d): %s",
		   fd, opt->desc->major, strerror(errno));
	    opt->desc = ODESC_ERROR; continue;
	 } else {
	    if (opt->value.u_bool) {
	       flag |= opt->desc->minor;
	    } else {
	       flag &= ~opt->desc->minor;
	    }
	    if (Fcntl_l(fd, opt->desc->major, flag) < 0) {
	       Error4("fcntl(%d, %d, %d): %s",
		      fd, opt->desc->major, flag, strerror(errno));
	       opt->desc = ODESC_ERROR; continue;
	    }
	 }

      } else if (opt->desc->func == OFUNC_IOCTL) {
	 if (Ioctl(fd, opt->desc->major, (void *)&opt->value) < 0) {
	    Error4("ioctl(%d, 0x%x, %p): %s",
		   fd, opt->desc->major, (void *)&opt->value, strerror(errno));
	    opt->desc = ODESC_ERROR; continue;
	 }

#if WITH_SOCKET
      } else if (opt->desc->func == OFUNC_SOCKOPT) {
#if 0 && STRUCT_LINGER
	 if (opt->desc->optcode == OPT_SO_LINGER) {
	    struct linger lingstru;
	    lingstru.l_onoff = (opt->value.u_int>=0 ? 1 : 0);
	    lingstru.l_linger = opt->value.u_int;
	    if (Setsockopt(fd, opt->desc->major, opt->desc->minor, &lingstru,
			   sizeof(lingstru)) < 0) {
	       Error6("setsockopt(%d, %d, %d, {%d,%d}, "F_Zu,
		      fd, opt->desc->major, opt->desc->minor, lingstru.l_onoff,
		      lingstru.l_linger, sizeof(lingstru));
	       opt->desc = ODESC_ERROR; continue;
	    }
	 } else {
#endif /* STRUCT_LINGER */
	    switch (opt->desc->type) {
	    case TYPE_BIN:
	       if (Setsockopt(fd, opt->desc->major, opt->desc->minor,
			      opt->value.u_bin.b_data, opt->value.u_bin.b_len)
		   < 0) {
		  Error6("setsockopt(%d, %d, %d, %p, %d): %s",
			 fd, opt->desc->major, opt->desc->minor,
			 opt->value.u_bin.b_data, opt->value.u_bin.b_len,
			 strerror(errno));
		  opt->desc = ODESC_ERROR; continue;
	       }
	       break;
	    case TYPE_BOOL:
	       if (Setsockopt(fd, opt->desc->major, opt->desc->minor,
			      &opt->value.u_bool, sizeof(opt->value.u_bool))
		   < 0) {
		  Error6("setsockopt(%d, %d, %d, {%d}, "F_Zu"): %s", fd,
			 opt->desc->major, opt->desc->minor,
			 opt->value.u_bool, sizeof(opt->value.u_bool),
			 strerror(errno));
		  opt->desc = ODESC_ERROR; continue;
	       }
	       break;
	    case TYPE_BYTE:
	       if (Setsockopt(fd, opt->desc->major, opt->desc->minor,
			      &opt->value.u_byte, sizeof(byte_t)) < 0) {
		  Error6("setsockopt(%d, %d, %d, {%u}, "F_Zu"): %s",
			 fd, opt->desc->major, opt->desc->minor,
			 opt->value.u_byte, sizeof(byte_t), strerror(errno));
		  opt->desc = ODESC_ERROR; continue;
	       }
	       break;
	    case TYPE_INT:
	       if (Setsockopt(fd, opt->desc->major, opt->desc->minor,
			      &opt->value.u_int, sizeof(int)) < 0) {
		  Error6("setsockopt(%d, %d, %d, {%d}, "F_Zu"): %s",
			 fd, opt->desc->major, opt->desc->minor,
			 opt->value.u_int, sizeof(int), strerror(errno));
		  opt->desc = ODESC_ERROR; continue;
	       }
	       break;
	    case TYPE_LONG:
	       if (Setsockopt(fd, opt->desc->major, opt->desc->minor,
			      &opt->value.u_long, sizeof(long)) < 0) {
		  Error6("setsockopt(%d, %d, %d, {%ld}, "F_Zu"): %s",
			 fd, opt->desc->major, opt->desc->minor,
			 opt->value.u_long, sizeof(long), strerror(errno));
		  opt->desc = ODESC_ERROR; continue;
	       }
	       break;
	    case TYPE_STRING:
	       if (Setsockopt(fd, opt->desc->major, opt->desc->minor,
			      opt->value.u_string,
			      strlen(opt->value.u_string)+1) < 0) {
		  Error6("setsockopt(%d, %d, %d, \"%s\", "F_Zu"): %s",
			 fd, opt->desc->major, opt->desc->minor,
			 opt->value.u_string, strlen(opt->value.u_string)+1,
			 strerror(errno));
		  opt->desc = ODESC_ERROR; continue;
	       }
	       break;
	    case TYPE_UINT:
	       if (Setsockopt(fd, opt->desc->major, opt->desc->minor,
			      &opt->value.u_uint, sizeof(unsigned int)) < 0) {
		  Error6("setsockopt(%d, %d, %d, {%u}, "F_Zu"): %s",
			 fd, opt->desc->major, opt->desc->minor,
			 opt->value.u_uint, sizeof(unsigned int),
			 strerror(errno));
		  opt->desc = ODESC_ERROR; continue;
	       }
	       break;
	    case TYPE_TIMEVAL:
	       if (Setsockopt(fd, opt->desc->major, opt->desc->minor,
			      &opt->value.u_timeval, sizeof(struct timeval)) < 0) {
		  Error7("setsockopt(%d, %d, %d, {%ld,%ld}, "F_Zu"): %s",
			 fd, opt->desc->major, opt->desc->minor,
			 opt->value.u_timeval.tv_sec, opt->value.u_timeval.tv_usec,
			 sizeof(struct timeval), strerror(errno));
		  opt->desc = ODESC_ERROR; continue;
	       }
	       break;
#if STRUCT_LINGER
	    case TYPE_LINGER:
	       {
		  struct linger lingstru;
		  lingstru.l_onoff = (opt->value.u_linger.l_onoff>=0 ? 1 : 0);
		  lingstru.l_linger = opt->value.u_linger.l_linger;
		  if (Setsockopt(fd, opt->desc->major, opt->desc->minor,
				 &lingstru, sizeof(lingstru)) < 0) {
		     Error6("setsockopt(%d, %d, %d, {%d,%d}): %s",
			    fd, opt->desc->major, opt->desc->minor,
			    lingstru.l_onoff, lingstru.l_linger,
			    strerror(errno));
		    opt->desc = ODESC_ERROR; continue;
		  }
	       }
	       break;
#endif /* STRUCT_LINGER */
	       /*! still many types missing; implement on demand */
#if !NDEBUG
	    default: Error1("applyopts(): type %d not implemented",
			    opt->desc->type);
	       opt->desc = ODESC_ERROR; continue;
#endif
	    }
#if 0 && STRUCT_LINGER
	 }
#endif /* STRUCT_LINGER */

      } else if (opt->desc->func == OFUNC_SOCKOPT_APPEND) {
	 switch (opt->desc->type) {
	    byte_t data[256];
	    size_t oldlen, newlen;
	 case TYPE_BIN:
	    oldlen = sizeof(data);
	    if (Getsockopt(fd, opt->desc->major, opt->desc->minor,
			   data, &oldlen)
		< 0) {
	       Error6("getsockopt(%d, %d, %d, %p, {F_Zu}): %s",
		      fd, opt->desc->major, opt->desc->minor, data, oldlen,
		      strerror(errno));
	       opt->desc = ODESC_ERROR; continue;
	    }
	    memcpy(&data[oldlen], opt->value.u_bin.b_data,
		   MIN(opt->value.u_bin.b_len, sizeof(data)-oldlen));
	    newlen = oldlen + MIN(opt->value.u_bin.b_len, sizeof(data)-oldlen);
	    if (Setsockopt(fd, opt->desc->major, opt->desc->minor,
			   data, newlen)
		< 0) {
	       Error6("setsockopt(%d, %d, %d, %p, %d): %s",
		      fd, opt->desc->major, opt->desc->minor, data, newlen,
		      strerror(errno));
	       opt->desc = ODESC_ERROR; continue;
	    }
	    break;
	 default:
	    Error2("internal: option \"%s\": unimplemented type %d",
		   opt->desc->defname, opt->desc->type);
	    break;
	 }
#endif /* WITH_SOCKET */
	 
#if HAVE_FLOCK
      } else if (opt->desc->func == OFUNC_FLOCK) {
	 if (Flock(fd, opt->desc->major) < 0) {
	    Error3("flock(%d, %d): %s",
		   fd, opt->desc->major, strerror(errno));
	    opt->desc = ODESC_ERROR; continue;
	 }
#endif /* defined(HAVE_FLOCK) */

      } else if (opt->desc->func == OFUNC_SPEC ||
		 opt->desc->func == OFUNC_FLAG) {
	 switch (opt->desc->optcode) {
	 case OPT_USER:
	 case OPT_USER_LATE:
	    if (Fchown(fd, opt->value.u_uidt, -1) < 0) {
	       Error3("fchown(%d, "F_uid", -1): %s",
		      fd, opt->value.u_uidt, strerror(errno));
	       opt->desc = ODESC_ERROR; continue;
	    }
	    break;
	 case OPT_GROUP:
	 case OPT_GROUP_LATE:
	    if (Fchown(fd, -1, opt->value.u_gidt) < 0) {
	       Error3("fchown(%d, -1, "F_gid"): %s",
		      fd, opt->value.u_gidt, strerror(errno));
	       opt->desc = ODESC_ERROR; continue;
	    }
	    break;
	 case OPT_PERM:
	 case OPT_PERM_LATE:
	    if (Fchmod(fd, opt->value.u_modet) < 0) {
	       Error3("fchmod(%d, %u): %s",
		      fd, opt->value.u_modet, strerror(errno));
	       opt->desc = ODESC_ERROR; continue;
	    }
	    break;
	 case OPT_FTRUNCATE:
	    if (Ftruncate(fd, opt->value.u_long) < 0) {
	       Error3("ftruncate(%d, %ld): %s",
		      fd, opt->value.u_long, strerror(errno));
	       opt->desc = ODESC_ERROR; continue;
	    }
	    break; 
	 case OPT_F_SETLK:
	 case OPT_F_SETLKW:
	    {
	       struct flock l;	/* Linux: <asm/fcntl.h> */
	       l.l_type   = F_WRLCK;
	       l.l_whence = SEEK_SET;
	       l.l_start  = 0;
	       l.l_len    = LONG_MAX;
	       l.l_pid    = 0;	/* hope this uses our current process */
	       if (Fcntl_lock(fd, opt->desc->major, &l) < 0) {
		  Error3("fcntl(%d, %d, {type=F_WRLCK,whence=SEEK_SET,start=0,len=LONG_MAX,pid=0}): %s", fd, opt->desc->major, strerror(errno));
		  opt->desc	 = ODESC_ERROR; continue;
	       }
	    }
	    break; 
	 case OPT_SETUID_EARLY:
	 case OPT_SETUID:
	    if (Setuid(opt->value.u_uidt) < 0) {
	       Error2("setuid("F_uid"): %s", opt->value.u_uidt,
		      strerror(errno));
	       opt->desc = ODESC_ERROR; continue;
	    }
	    break;
	 case OPT_SETGID_EARLY:
	 case OPT_SETGID:
	    if (Setgid(opt->value.u_gidt) < 0) {
	       Error2("setgid("F_gid"): %s", opt->value.u_gidt,
		      strerror(errno));
	       opt->desc = ODESC_ERROR; continue;
	    }
	    break;
	 case OPT_SUBSTUSER:
	    {
	       struct passwd *pwd;
	       if ((pwd = getpwuid(opt->value.u_uidt)) == NULL) {
		  Error1("getpwuid("F_uid"): no such user",
			 opt->value.u_uidt);
		  opt->desc = ODESC_ERROR; continue;
	       }
	       if (Initgroups(pwd->pw_name, pwd->pw_gid) < 0) {
		  Error3("initgroups(%s, "F_gid"): %s",
			 pwd->pw_name, pwd->pw_gid, strerror(errno));
		  opt->desc = ODESC_ERROR; continue;
	       }
	       if (Setgid(pwd->pw_gid) < 0) {
		  Error2("setgid("F_gid"): %s", pwd->pw_gid,
			 strerror(errno));
		  opt->desc = ODESC_ERROR; continue;
	       }
	       if (Setuid(opt->value.u_uidt) < 0) {
		  Error2("setuid("F_uid"): %s", opt->value.u_uidt,
			 strerror(errno));
		  opt->desc = ODESC_ERROR; continue;
	       }
#if 1
	       if (setenv("USER", pwd->pw_name, 1) < 0)
		  Error1("setenv(\"USER\", \"%s\", 1): insufficient space",
			 pwd->pw_name);
	       if (setenv("LOGNAME", pwd->pw_name, 1) < 0)
		  Error1("setenv(\"LOGNAME\", \"%s\", 1): insufficient space",
			 pwd->pw_name);
	       if (setenv("HOME", pwd->pw_dir, 1) < 0)
		  Error1("setenv(\"HOME\", \"%s\", 1): insufficient space",
			 pwd->pw_dir);
	       if (setenv("SHELL", pwd->pw_shell, 1) < 0)
		  Error1("setenv(\"SHELL\", \"%s\", 1): insufficient space",
			 pwd->pw_shell);
#endif
	    }
	    break;
	 case OPT_SUBSTUSER_DELAYED:
	    {
	       struct passwd *pwd;

	       if ((pwd = getpwuid(opt->value.u_uidt)) == NULL) {
		  Error1("getpwuid("F_uid"): no such user",
			 opt->value.u_uidt);
		  opt->desc = ODESC_ERROR; continue;
	       }
	       delayeduser_uid = opt->value.u_uidt;
	       delayeduser_gid = pwd->pw_gid;
	       if ((delayeduser_name = strdup(pwd->pw_name)) == NULL) {
		  Error1("strdup("F_Zu"): out of memory",
			 strlen(pwd->pw_name)+1);
		  opt->desc = ODESC_ERROR; continue;
	       }
	       if ((delayeduser_dir = strdup(pwd->pw_dir)) == NULL) {
		  Error1("strdup("F_Zu"): out of memory",
			 strlen(pwd->pw_dir)+1);
		  opt->desc = ODESC_ERROR; continue;
	       }
	       if ((delayeduser_shell = strdup(pwd->pw_shell)) == NULL) {
		  Error1("strdup("F_Zu"): out of memory",
			 strlen(pwd->pw_shell)+1);
		  opt->desc = ODESC_ERROR; continue;
	       }
	       /* function to get all supplementary groups of user */
	       delayeduser_ngids = sizeof(delayeduser_gids)/sizeof(gid_t);
	       getusergroups(delayeduser_name, delayeduser_gids,
			     &delayeduser_ngids);
	       delayeduser = true;
	    }
	    break;
	 case OPT_CHROOT_EARLY:
	 case OPT_CHROOT:
	    if (Chroot(opt->value.u_string) < 0) {
	       Error2("chroot(\"%s\"): %s", opt->value.u_string,
		      strerror(errno));
	       opt->desc = ODESC_ERROR; continue;
	    }
	    if (Chdir("/") < 0) {
	       Error1("chdir(\"/\"): %s", strerror(errno));
	    }
	    break;
	 case OPT_SETSID:
	    if (Setsid() < 0) {
	       Warn1("setsid(): %s", strerror(errno));
	       if (Setpgid(getpid(), getppid()) < 0) {
		  Warn3("setpgid(%d, %d): %s",
			getpid(), getppid(), strerror(errno));
	       } else {
		  if (Setsid() < 0) {
		     Error1("setsid(): %s", strerror(errno));
		  }		  
	       }
	    }
	    break;
	 case OPT_SETPGID:
	    if (Setpgid(0, opt->value.u_int) < 0) {
	       Warn2("setpgid(0, "F_pid"): %s",
		     opt->value.u_int, strerror(errno));
	    }
	    break;
	 case OPT_TIOCSCTTY:
	    {
	       int mytty;
	       /* this code idea taken from ssh/pty.c: make pty controlling term. */
	       if ((mytty = Open("/dev/tty", O_NOCTTY, 0640)) < 0) {
		  Warn1("open(\"/dev/tty\", O_NOCTTY, 0640): %s", strerror(errno));
	       } else {
#ifdef TIOCNOTTY
		  if (Ioctl(mytty, TIOCNOTTY, NULL) < 0) {
		     Warn2("ioctl(%d, TIOCNOTTY, NULL): %s",
			   mytty, strerror(errno));
		  }
#endif
		  if (Close(mytty) < 0) {
		     Warn2("close(%d): %s",
			   mytty, strerror(errno));
		  }
	       }
#ifdef TIOCSCTTY
	       if (Ioctl(fd, TIOCSCTTY, NULL) < 0) {
		  Warn2("ioctl(%d, TIOCSCTTY, NULL): %s", fd, strerror(errno));
	       }
#endif
	       if (Tcsetpgrp(0, getpid()) < 0) {
		  Warn2("tcsetpgrp("F_pid"): %s", getpid(), strerror(errno));
	       }
	    }
	    break;
	 default: Error1("applyopts(): option \"%s\" not implemented",
			 opt->desc->defname);
	    opt->desc = ODESC_ERROR; continue;
	 }

#if WITH_TERMIOS
      } else if (opt->desc->func == OFUNC_TERMIOS_FLAG) {
#if 0
	 union {
	    struct termios termarg;
	    tcflag_t flags[4];
	 } tdata;
	 if (Tcgetattr(fd, &tdata.termarg) < 0) {
	    Error3("tcgetattr(%d, %p): %s",
		   fd, &tdata.termarg, strerror(errno));
	    opt->desc = ODESC_ERROR; continue;
	 }
	 if (opt->value.u_bool) {
	    tdata.flags[opt->desc->major] |= opt->desc->minor;
	 } else {
	    tdata.flags[opt->desc->major] &= ~opt->desc->minor;
	 }
	 if (Tcsetattr(fd, TCSADRAIN, &tdata.termarg) < 0) {
	    Error3("tcsetattr(%d, TCSADRAIN, %p): %s",
		   fd, &tdata.termarg, strerror(errno));
	    opt->desc = ODESC_ERROR; continue;
	 }
#else
	 if (xiotermiosflag_applyopt(fd, opt) < 0) {
	    continue;
	 }
#endif

      } else if (opt->desc->func == OFUNC_TERMIOS_VALUE) {
	 union {
	    struct termios termarg;
	    tcflag_t flags[4];
	 } tdata;
	 if (((opt->value.u_uint << opt->desc->arg3) & opt->desc->minor) !=
	     (opt->value.u_uint << opt->desc->arg3)) {
	    Error2("option %s: invalid value %u",
		   opt->desc->defname, opt->value.u_uint);
	    opt->desc = ODESC_ERROR; continue;
	 }
	 if (Tcgetattr(fd, &tdata.termarg) < 0) {
	    Error3("tcgetattr(%d, %p): %s",
		   fd, &tdata.termarg, strerror(errno));
	    opt->desc = ODESC_ERROR; continue;
	 }

	 tdata.flags[opt->desc->major] &= ~opt->desc->minor;
	 tdata.flags[opt->desc->major] |=
	    ((opt->value.u_uint << opt->desc->arg3) & opt->desc->minor);
	 if (Tcsetattr(fd, TCSADRAIN, &tdata.termarg) < 0) {
	    Error3("tcsetattr(%d, TCSADRAIN, %p): %s",
		   fd, &tdata.termarg, strerror(errno));
	    opt->desc = ODESC_ERROR; continue;
	 }

      } else if (opt->desc->func == OFUNC_TERMIOS_PATTERN) {
	 union {
	    struct termios termarg;
	    tcflag_t flags[4];
	 } tdata;
	 if (Tcgetattr(fd, &tdata.termarg) < 0) {
	    Error3("tcgetattr(%d, %p): %s",
		   fd, &tdata.termarg, strerror(errno));
	    opt->desc = ODESC_ERROR; continue;
	 }
	 tdata.flags[opt->desc->major] &= ~opt->desc->arg3;
	 tdata.flags[opt->desc->major] |= opt->desc->minor;
	 if (Tcsetattr(fd, TCSADRAIN, &tdata.termarg) < 0) {
	    Error3("tcsetattr(%d, TCSADRAIN, %p): %s",
		   fd, &tdata.termarg, strerror(errno));
	    opt->desc = ODESC_ERROR; continue;
	 }

      } else if (opt->desc->func == OFUNC_TERMIOS_CHAR) {
	 struct termios termarg;
	 if (Tcgetattr(fd, &termarg) < 0) {
	    Error3("tcgetattr(%d, %p): %s",
		   fd, &termarg, strerror(errno));
	    opt->desc = ODESC_ERROR; continue;
	 }
	 termarg.c_cc[opt->desc->major] = opt->value.u_byte;
	 if (Tcsetattr(fd, TCSADRAIN, &termarg) < 0) {
	    Error3("tcsetattr(%d, TCSADRAIN, %p): %s",
		   fd, &termarg, strerror(errno));
	    opt->desc = ODESC_ERROR; continue;
	 }

#ifdef HAVE_TERMIOS_ISPEED
      } else if (opt->desc->func == OFUNC_TERMIOS_SPEED) {
	 union {
	    struct termios termarg;
	    speed_t speeds[sizeof(struct termios)/sizeof(speed_t)];
	 } tdata;
	 if (Tcgetattr(fd, &tdata.termarg) < 0) {
	    Error3("tcgetattr(%d, %p): %s",
		   fd, &tdata.termarg, strerror(errno));
	    opt->desc = ODESC_ERROR; continue;
	 }
	 tdata.speeds[opt->desc->major] = opt->value.u_uint;
	 if (Tcsetattr(fd, TCSADRAIN, &tdata.termarg) < 0) {
	    Error3("tcsetattr(%d, TCSADRAIN, %p): %s",
		   fd, &tdata.termarg, strerror(errno));
	    opt->desc = ODESC_ERROR; continue;
	 }
#endif /* HAVE_TERMIOS_ISPEED */

      } else if (opt->desc->func == OFUNC_TERMIOS_SPEC) {
	 struct termios termarg;
	 if (Tcgetattr(fd, &termarg) < 0) {
	    Error3("tcgetattr(%d, %p): %s",
		   fd, &termarg, strerror(errno));
	    opt->desc = ODESC_ERROR; continue;
	 }
	 switch (opt->desc->optcode) {
	 case OPT_RAW:
	    termarg.c_iflag &=
	      ~(IGNBRK|BRKINT|IGNPAR|PARMRK|INPCK|ISTRIP|INLCR|IGNCR|ICRNL|IXON|IXOFF
#ifdef IUCLC
		|IUCLC
#endif
		|IXANY|IMAXBEL);
	    termarg.c_iflag |= (0);
	    termarg.c_oflag &= ~(OPOST);
	    termarg.c_oflag |= (0);
	    termarg.c_cflag &= ~(0);
	    termarg.c_cflag |= (0);
	    termarg.c_lflag &= ~(ISIG|ICANON
#ifdef XCASE
				 |XCASE
#endif
				 );
	    termarg.c_lflag |= (0);
	    termarg.c_cc[VMIN] = 1;
	    termarg.c_cc[VTIME] = 0;
	    break;
	 case OPT_SANE:
	    /* cread -ignbrk brkint  -inlcr  -igncr  icrnl
              -ixoff  -iuclc  -ixany  imaxbel opost -olcuc -ocrnl
              onlcr -onocr -onlret -ofill -ofdel nl0 cr0 tab0 bs0
              vt0 ff0 isig icanon iexten echo echoe echok -echonl
              -noflsh -xcase -tostop -echoprt echoctl echoke, and
              also  sets  all special characters to their default
              values.
*/
	    termarg.c_iflag &= ~(IGNBRK|INLCR|IGNCR|IXOFF
#ifdef IUCLC
				 |IUCLC
#endif
				 |IXANY);
	    termarg.c_iflag |= (BRKINT|ICRNL|IMAXBEL);
	    termarg.c_oflag &= ~(0	/* for canonical reasons */
#ifdef OLCUC
				 |OLCUC
#endif
#ifdef OCRNL
				 |OCRNL
#endif
#ifdef ONOCR
				 |ONOCR
#endif
#ifdef ONLRET
				 |ONLRET
#endif
#ifdef OFILL
				 |OFILL
#endif
#ifdef OFDEL
				 |OFDEL
#endif
#ifdef NLDLY
				 |NLDLY
#endif
#ifdef CRDLY
				 |CRDLY
#endif
#ifdef TABDLY
				 |TABDLY
#endif
#ifdef BSDLY
				 |BSDLY
#endif
#ifdef VTDLY
				 |VTDLY
#endif
#ifdef FFDLY
				 |FFDLY
#endif
				 );
	    termarg.c_oflag |= (OPOST|ONLCR
#ifdef NL0
				|NL0
#endif
#ifdef CR0
				|CR0
#endif
#ifdef TAB0
				|TAB0
#endif
#ifdef BS0
				|BS0
#endif
#ifdef VT0
				|VT0
#endif
#ifdef FF0
				|FF0
#endif
				);
	    termarg.c_cflag &= ~(0);
	    termarg.c_cflag |= (CREAD);
	    termarg.c_lflag &= ~(ECHONL|NOFLSH
#ifdef XCASE
				 |XCASE
#endif
				 |TOSTOP
#ifdef ECHOPRT
				 |ECHOPRT
#endif
				 );
	    termarg.c_lflag |= (ISIG|ICANON|IEXTEN|ECHO|ECHOE|ECHOK|ECHOCTL|ECHOKE);
	    /*! "sets characters to their default values... - which? */
	    break;
	 default:
	    Error("TERMIOS option not handled - internal error?");
	 }
	 if (Tcsetattr(fd, TCSADRAIN, &termarg) < 0) {
	    Error3("tcsetattr(%d, TCSADRAIN, %p): %s",
		   fd, &termarg, strerror(errno));
	    opt->desc = ODESC_ERROR; continue;
	 }

#endif /* WITH_TERMIOS */
      } else {
	/*Error1("applyopts(): function %d not implemented",
	  opt->desc->func);*/
	 if (opt->desc->func != OFUNC_EXT) {
	    Error1("applyopts(): option \"%s\" does not apply",
		   opt->desc->defname);
	    opt->desc = ODESC_ERROR;
	    ++opt;
	    continue;
	 }
	 ++opt;
	 continue;
      }
      opt->desc = ODESC_DONE;
      ++opt;
   }
   return 0;
}

/* applies to fd all options belonging to phases */
/* note: not all options can be applied this way (e.g. OFUNC_SPEC with PH_OPEN)
   implemented are: OFUNC_FCNTL, OFUNC_SOCKOPT (probably not all types),
   OFUNC_TERMIOS_FLAG, OFUNC_TERMIOS_PATTERN, and some OFUNC_SPEC */
int applyopts2(int fd, struct opt *opts, unsigned int from, unsigned int to) {
   unsigned int i;
   int stat;

   for (i = from; i <= to; ++i) {
      if ((stat = applyopts(fd, opts, i)) < 0)
	 return stat;
   }
   return 0;
}

/* apply and consume all options of type FLAG and group.
   Return 0 if everything went right, or -1 if an error occurred. */
int applyopts_flags(struct opt *opts, int group, flags_t *result) {
   struct opt *opt = opts;

   if (!opts)  return 0;

   while (opt->desc != ODESC_END) {
      if (opt->desc != ODESC_DONE && 
	  (opt->desc->group & group)) {
	 if (opt->desc->func == OFUNC_FLAG) {
	    if (opt->value.u_bool) {
	       *result |= opt->desc->major;
	    } else {
	       *result &= ~opt->desc->major;
	    }
	    opt->desc = ODESC_DONE;
	 } else if (opt->desc->func == OFUNC_FLAG_PATTERN) {
	    *result &= ~opt->desc->minor;
	    *result |= opt->desc->major;
	    opt->desc = ODESC_DONE;
	 }
      }
      ++opt;
   }
   return 0;
}



/* set the FD_CLOEXEC fcntl if the options do not set it to 0 */
int applyopts_cloexec(int fd, struct opt *opts) {
   bool docloexec = 1;

   if (!opts)  return 0;

   retropt_bool(opts, OPT_CLOEXEC, &docloexec);
   if (docloexec) {
      if (Fcntl_l(fd, F_SETFD, FD_CLOEXEC) < 0) {
	 Warn2("fcntl(%d, F_SETFD, FD_CLOEXEC): %s", fd, strerror(errno));
      }
   }
   return 0;
}

#if 0
/* apply options of 'early' phase, groups PROCESS and NAMED. path should exist
   for NAMED (not checked here). */
int applyopts_early(const char *path, struct opt *opts) {
   uid_t uid = -1;
   gid_t gid = -1;
   mode_t perm = 0666;

   if (retropt_uidt(opts, OPT_SETUID, &uid) >= 0) {
     Setuid()}
   retropt_uidt(opts, OPT_SETGID, &gid);
   /*retropt_uidt(opts, OPT_USER, &user);*/
   retropt_gidt(opts, OPT_GROUP_EARLY, &group);
   /*retropt_gidt(opts, OPT_GROUP, &group);*/

   if (uid != (uid_t)-1 || gid != (gid_t)-1) {
      if (Chown(path, user, group) < 0) {
	 Error4("chown(\"%s\", "F_uid", "F_gid"): %s", path, user, group,
		strerror(errno));
      }
   }

   if (retropt_modet(opts, OPT_PERM_EARLY, &perm) >= 0 /*||
       retropt_modet(opts, OPT_PERM, &perm) >= 0*/) {
      if (Chmod(path, perm) < 0) {
	 Error3("chmod(\"%s\", "F_mode"): %s", path, perm, strerror(errno));
      }
   }

   return 0;
}
#endif /* 0 */

int applyopts_fchown(int fd, struct opt *opts) {
   uid_t user = -1;
   gid_t group = -1;

   retropt_uidt(opts, OPT_USER, &user);
   retropt_gidt(opts, OPT_GROUP, &group);

   if (user != (uid_t)-1 || group != (gid_t)-1) {
      if (Fchown(fd, user, group) < 0) {
	 Error4("fchown(%d, "F_uid", "F_gid"): %s", fd, user, group,
		strerror(errno));
	 return STAT_RETRYLATER;
      }
   }
   return 0;
}

int applyopts_offset(xiofile_t *xfd, struct opt *opts) {
   struct opt *opt;
   unsigned char *ptr;

   opt = opts; while (opt->desc != ODESC_END) {
      if ((opt->desc == ODESC_DONE) ||
	  opt->desc->func != OFUNC_OFFSET)  {
	 ++opt; continue; }

      ptr = (unsigned char *)xfd + opts->desc->major;
      switch (opt->desc->type) {
      case TYPE_DOUBLE:
	 *(double *)ptr = opts->value.u_double;  break;
      case TYPE_STRING:
	 if ((*(char **)ptr = strdup(opts->value.u_string)) == NULL) {
	    Error1("strdup("F_Zu"): out of memory",
		   strlen(opts->value.u_string)+1);
	 }
 	 break;
      default:
	 Error1("applyopts_offset(): type %d not implemented",
	       opt->desc->type);
      }
      opt->desc = ODESC_DONE;
      ++opt;
   }
   return 0;
}

/* applies to xfd all OFUNC_EXT options belonging to phase */
int applyopts_single(struct single *fd, struct opt *opts, enum e_phase phase) {
   struct opt *opt;

   if (!opts)  return 0;

   opt = opts; while (opt->desc != ODESC_END) {
      if ((opt->desc == ODESC_DONE) ||
	  (opt->desc->phase != phase && phase != PH_ALL) ||
	  opt->desc->func != OFUNC_EXT)  {
	 ++opt; continue; }
      switch (opt->desc->optcode) {
      case OPT_IGNOREEOF:
	 fd->ignoreeof = true;
	 opt->desc = ODESC_DONE; break;
      case OPT_CR:
	 fd->lineterm = LINETERM_CR;
	 opt->desc = ODESC_DONE; break;
      case OPT_CRNL:
	 fd->lineterm = LINETERM_CRNL;
	 opt->desc = ODESC_DONE; break;
#if !NDEBUG
      default:
	 Error1("applyopts_single(): optioncode %d not implemented",
	       opt->desc->optcode);
#endif
      }
      ++opt;
   }
   return 0;
}

/* apply remaining options to file descriptor, and tell us if something is
   still unused */
int _xio_openlate(struct single *fd, struct opt *opts) {
   int numleft;
   int result;

   _xioopen_setdelayeduser();

   if ((result = applyopts(fd->fd, opts, PH_LATE)) < 0) {
      return result;
   }
   if ((result = applyopts_single(fd, opts, PH_LATE)) < 0) {
      return result;
   }
   if ((result = applyopts(fd->fd, opts, PH_LATE2)) < 0) {
      return result;
   }

   if ((numleft = leftopts(opts)) > 0) {
      Error1("%d option(s) could not be used", numleft);
      showleft(opts);
      return STAT_NORETRY;
   }
   return 0;
}

int dropopts(struct opt *opts, unsigned int phase) {
   struct opt *opt;

   opt = opts; while (opt && opt->desc != ODESC_END) {
      if (opt->desc != ODESC_DONE && opt->desc->phase == phase) {
	 Debug1("ignoring option \"%s\"", opt->desc->defname);
	 opt->desc = ODESC_DONE;
      }
      ++opt;
   }
   return 0;
}

int dropopts2(struct opt *opts, unsigned int from, unsigned int to) {
   unsigned int i;

   for (i = from; i <= to; ++i) {
      dropopts(opts, i);
   }
   return 0;
}

